# -*- coding: utf-8 -*-
"""
Script Name: Logik Portal
Script Version: 5.9.1
Flame Version: 2022
Written by: Michael Vaglienty
Crying Croc Design by: Enid Dalkoff
Creation Date: 10.31.20
Update Date: 04.18.24

Custom Action Type: Flame Main Menu

Description:

    Share/install python scripts, batch setups, archives, and download matchboxes

Menu:

    Flame Main Menu -> Logik -> Logik Portal

To install:

    Copy script into /opt/Autodesk/shared/python/logik_portal

Updates:

    v5.9.1 04.18.24

        Fixed: Install button not properly working for python scripts. Thanks for catching Mottizle!

    v5.9.0 03.05.24

        Added column to python script tab to show minimum and maximum flame versions required to run script.
        Scripts that require a newer or older version of flame will be greyed out in the list and not installable.

    v5.8.2 02.08.24

        Python scripts can now sorted by year when clicking on year header.
        Dates are flipped from dd.mm.yy to yy.mm.dd for sorting.

        Replaced browse buttons in submit windows with clickable line edit file browsers.

    v5.8.1 01.21.24

        Fixed: Submit buttons not working after submitting script unless portal is restarted.

    v5.8.0 01.18.24

        Updates to UI/PySide.

    v5.7.1 10.09.23

        Fixed progress window overflow error when uploading/downloading large files.

    v5.7.0 10.03.23

        Progress windows added when downloading or uploading files.

        When uploading python scripts __pycache__ folder is now excluded from tar file.

        Updated to pyflame lib v2.

    v5.6.1 06.26.23

        Updated script versioning to semantic versioning.

        Removed old version check of python script uploads to allow for semantic versioning.

        Updated password window for uploading python scripts.

        Main tabs now longer have outline when selected in linux.

    v5.6 03.02.23

        Updated FTP server info.

    v5.5 02.04.23

        Added search to matchbox tab.

    v5.4 02.02.23

        Improvements to Matchbox tab:

            All current matchboxes are now listed with descriptions.

            Matchboxes can be sorted by Name, Shader Type, and Author.

            Matchboxes can be added to current batch setup.

        Added check to make sure script is installed to correct path.

        Portal now opens to last used tab.

        Portal updates will be shown on whatever tab script first opens to.

    v5.3 01.25.23

        Matchboxes will now install into a directory called LOGIK in the selected directory.

        Reverted menu for Flame 2023.2+ to Flame Main Menu -> Logik -> Logik Portal. Scripts downloaded from the Portal that have a setup menu
        in the future will have their menu added under Flame Main Menu -> Logik -> Logik Portal Setup -> Script Name for clarity.

        Updated config file loading/saving.

    v5.2 01.08.23

        Updates to the Logik Portal are now shown in the main window when the script first loads up.

    v5.1 12.22.22

        Fixed possible ssl error when downloading matchboxes.

    v5.0 11.28.22

        Updated with new FTP server.

        Autodesk python scripts provided with Flame 2023.2+ are now listed/installable through the Portal.

        Maximum archive size increased to 1GB.

    v4.2 09.06.22

        Updated menu for Flame 2023.2+:

            Flame Main Menu -> Logik Portal

    v4.1 07.22.22

        Messages print to Flame message window - Flame 2023.1+.

        Added Flame file browser - Flame 2023.1+.

        pyflame_lib files aren't shown in the installed scripts list anymore.

        Matchbox install path now defaults to  /opt/Autodesk/presets/FLAME_VERSION/matchbox/shaders

    v4.0 03.23.22

        Updated UI for Flame 2023.

        Moved UI widgets to external file.

    v3.0 12.09.21

        Getting Flame version is updated to work with new PR versioning.

        Moved a few buttons around.

    v2.9 12.02.21

        Python script upload login bug fix.

    v2.8 11.17.21

        Login info for uploading scripts only needs to be entered first time something is uploaded.

    v2.7 10.16.21

        Install Local button added to python tab to install python scripts from local drive.

        Improved Flame version detection.

        Script will now attempt to download matchbox collection from website. If website is down, it will download from portal ftp.

    v2.6 09.06.21

        Misc bug fixes / fixed problem with not being able to enter system password to load matchboxes to write protected folder.

    v2.5 07.30.21

        Added ability to upload/download archives - Archive size limit is 200MB.

        Config is now XML.

    v2.4 07.23.21

        Added python submit button back. User name and password now required to submit scripts.

        Fixed bug - files starting with . sometimes caused script to not work.

    v2.3 07.06.21

        Added Logik Matchbox archive to Portal FTP. Matchbox archive now stored on FTP instead of pulling directly from logik-matchbook.org.

    v2.2 06.03.21

        Updated to be compatible with Flame 2022/Python 3.7.

        Removed python script submission ability. Scripts can now be added through github submissions only.

    v1.6 03.14.21

        UI improvements/updates - UI elements to classes.

        Added contextual menus to python tab to install and delete scripts and to batch tab to download batch setups.

        User will be prompted for system password when trying to download matchboxes to protected folders such as /opt/Autodesk/presets/2021.1/matchbox/shaders.

        If newer version of installed script is available on portal it will be highlighted in portal list.

        If newer version of flame is required for a script, script entry will be greyed out.

        If newer version of flame is required for a batch setup, batch setup entry will be greyed out.

        Batch setups now properly download into paths with spaces in folder names.

        User will get message if script folder needs permissions changed to create temp folders/files.

        File browse buttons removed - browser now opens when clicking lineedit field.

        If new version of a python script is submitted old script will be removed.

    v1.5 02.27.21

        UI code updates.

        Fixed bug causing script to hang when reading descriptions on certain scripts.

        Fixed batch submit button.

    v1.4 01.25.21

        Fixed temp path for logik matchbox install.

    v1.3 01.14.21

        Script description info can now be entered in Portal UI instead of being in script header.

        Fixed font size for linux.

    v1.2 12.29.20

        Fixed problems with script running on Flame with extra .x in Flame version.
"""

# ---------------------------------------- #
# Imports

import datetime
import os
import re
import shutil
import xml.etree.ElementTree as ET
from ftplib import FTP
from functools import partial
from subprocess import PIPE, Popen

import flame
from pyflame_lib_logik_portal import *

# Try to import PySide6, otherwise import PySide2
try:
    from PySide6 import QtCore, QtGui, QtWidgets
except ImportError:
    from PySide2 import QtCore, QtGui, QtWidgets

# ---------------------------------------- #
# Main Script

SCRIPT_NAME = 'Logik Portal'
SCRIPT_VERSION = 'v5.9.1'
SCRIPT_PATH = '/opt/Autodesk/shared/python/logik_portal'

LOGIK_FTP = '45.79.19.175'

class LogikPortal():

    def __init__(self, selection):

        print('\n')
        print('>' * 10, f'{SCRIPT_NAME} {SCRIPT_VERSION}', '<' * 10, '\n')

        # Check script path
        if os.path.dirname(os.path.abspath(__file__)) != SCRIPT_PATH:
            PyFlameMessageWindow(
                message='Script path is incorrect. Please reinstall script.<br><br>Script path should be:<br><br>' + SCRIPT_PATH,
                script_name=SCRIPT_NAME,
                type=MessageType.ERROR,
                )
            return

        # Define paths
        self.shared_script_path = '/opt/Autodesk/shared/python'

        # Get version of flame
        self.flame_full_version = flame.get_version()
        self.flame_version = pyflame.get_flame_version()

        # Set Autodesk python scripts path
        self.autodesk_scripts_path = f'/opt/Autodesk/flame_{self.flame_full_version}/python_utilities/scripts'
        if not os.path.isdir(self.autodesk_scripts_path):
            self.autodesk_scripts_path = f'/opt/Autodesk/flare_{self.flame_full_version}/python_utilities/scripts'
        #print('autodesk_scripts_path:', self.autodesk_scripts_path)

        # Set install path for local python scripts
        if float(self.flame_version) >= 2023.2:
            install_local_path = self.autodesk_scripts_path
        else:
            install_local_path = '/opt/Autodesk'
        #print('install_local_path:', install_local_path)

        # Load config file
        self.settings = PyFlameConfig(
            script_name=SCRIPT_NAME,
            script_path=SCRIPT_PATH,
            config_values={
                'python_submit_all_files': 'False',
                'matchbox_path': f'/opt/Autodesk/presets/{self.flame_full_version}/matchbox/shaders',
                'batch_setup_download_path': '/opt/Autodesk',
                'batch_submit_path': '/opt/Autodesk',
                'script_submit_path': '/opt/Autodesk',
                'script_install_local_path': f'{install_local_path}',
                'open_batch': 'False',
                'archive_download_path': '/opt/Autodesk',
                'archive_submit_path': '/opt/Autodesk',
                'username': '',
                'password': '',
                'last_tab': '0',
                }
            )

        # Get user
        self.flame_current_user = flame.users.current_user.name

        # Create temp folder
        self.temp_folder = os.path.join(SCRIPT_PATH, 'temp')
        if not os.path.isdir(self.temp_folder):
            try:
                os.makedirs(self.temp_folder)
            except:
                PyFlameMessageWindow(
                    message=f'{SCRIPT_NAME}: Script needs full permissions to script folder.<br><br>In shell/terminal type:<br><br>chmod 777 /opt/Autodesk/shared/python/logik_portal<br>',
                    script_name=SCRIPT_NAME,
                    type=MessageType.ERROR,
                )
                return

        # Check internet connection to ftp
        try:
            self.ftp_download_connect()
        except:
            PyFlameMessageWindow(
                message="Can't connect to Logik Portal.<br>Check internet connection and try again.",
                script_name=SCRIPT_NAME,
                type=MessageType.ERROR,
            )
            return

        #  Init variables
        self.ftp_script_list = []
        self.installed_script_dict = {}
        self.file_description = ''
        self.tar_path = ''
        self.tar_file_name = ''
        self.batch_setups_xml_path = ''
        self.python_scripts_xml_path = ''
        self.sudo_password = ''
        self.updates = ''

        self.main_window()
        self.get_installed_scripts(self.installed_scripts_tree, self.shared_script_path)
        self.get_autodesk_scripts()

        self.download_xmls()

        self.update_matchbox_tree()
        self.get_batch_setups()
        self.get_archives()
        self.get_ftp_scripts()

        # Check first items in tree lists for install/download button disable
        self.check_archive_flame_version(self.archives_tree, 0)

        self.check_batch_flame_version(self.batch_setups_tree, 0)

        self.check_script_flame_version(self.portal_scripts_tree, 0)

        # Go to last used tab
        self.main_tabs.setCurrentIndex(self.settings.last_tab)

        # Get updates file from FTP
        self.get_updates()

        # Close ftp connection
        self.ftp.quit()

        pyflame.message_print(
            message='Welcome to the Logik Portal.',
            script_name=SCRIPT_NAME,
            )

    # ----------------------------------------------------------------- #

    def ftp_download_connect(self):

        # Connect to ftp
        self.ftp = FTP(f'{LOGIK_FTP}')
        self.ftp.login('logik_portal_download', 'L0gikD0wnL0ad#20')

        pyflame.message_print(
            message='Connected to Logik Portal.',
            script_name=SCRIPT_NAME,
            )

    def ftp_upload_connect(self):

        # Connect to ftp
        self.ftp = FTP(f'{LOGIK_FTP}')
        self.ftp.login('logik_portal_upload', 'L0gikUpl0ad#20')

        pyflame.message_print(
            message='Connected to Logik Portal.',
            script_name=SCRIPT_NAME,
            )

    def ftp_disconnect(self):

        self.ftp.quit()

        pyflame.message_print(
            message='Disconnected from Logik Portal.',
            script_name=SCRIPT_NAME,
            )

    # ----------------------------------------------------------------- #

    def main_window(self):

        self.window = PyFlameWindow(
            width=1200,
            height=750,
            title=f'{SCRIPT_NAME} <small>{SCRIPT_VERSION}'
            )

        # Tabs
        self.main_tabs = QtWidgets.QTabWidget()

        self.window.tab1 = QtWidgets.QWidget()
        self.window.tab2 = QtWidgets.QWidget()
        self.window.tab3 = QtWidgets.QWidget()
        self.window.tab4 = QtWidgets.QWidget()
        self.window.tab5 = QtWidgets.QWidget()

        self.main_tabs.addTab(self.window.tab1, 'Python Scripts')
        self.main_tabs.addTab(self.window.tab3, 'Matchbox')
        self.main_tabs.addTab(self.window.tab4, 'Batch Setups')
        self.main_tabs.addTab(self.window.tab5, 'Archives')

        self.python_scripts_tab()
        self.matchbox_tab()
        self.batch_setups_tab()
        self.archives_tab()

        self.main_tabs.setFocusPolicy(QtCore.Qt.NoFocus)

        # UI Widget Layout
        vbox = QtWidgets.QVBoxLayout()
        vbox.addWidget(self.main_tabs)

        self.window.add_layout(vbox)

        self.window.show()

        return self.window

    def python_scripts_tab(self):

        def login_check():

            def check_login(username, password):

                # Try connecting to ftp
                try:
                    ftp = FTP(f'{LOGIK_FTP}')
                    ftp.login(username, password)
                    ftp.cwd('/')

                    self.settings.save_config(
                        script_name=SCRIPT_NAME,
                        script_path=SCRIPT_PATH,
                        config_values={
                            'username': username,
                            'password': password
                            }
                        )

                    submit_script()
                except:
                    PyFlameMessageWindow(
                        message='Login incorrect, try again.',
                        script_name=SCRIPT_NAME,
                        type=MessageType.ERROR,
                        )
                    return

            if self.settings.username and self.settings.password:
                submit_script()
            else:
                password_window = PyFlamePasswordWindow(message='Portal login required to submit a python script.', title='Python Submit Login', user_name_prompt=True)
                username, password = password_window.username_password()
                if not username or not password:
                    return

                if username and password:
                    check_login(username, password)

        def submit_script():

            def upload_script():

                def upload():

                    def save_config():
                        print('button checked1:', self.settings.python_submit_all_files)
                        # Save path to config file
                        self.settings.save_config(
                            script_name=SCRIPT_NAME,
                            script_path=SCRIPT_PATH,
                            config_values={
                                'script_submit_path': self.submit_script_path_entry.text(),
                                'python_submit_all_files': self.all_files_btn.isChecked(),
                                }
                            )
                        print('button checked2:', self.settings.python_submit_all_files)
                        self.settings.all_files_btn = False
                        print('self.settings.python_submit_all_files:', self.settings.python_submit_all_files)

                    def create_script_xml(self):

                        description_text = self.submit_script_description_text_edit.toPlainText()
                        description_text = description_text.replace("'", "\"")
                        description_text = description_text.replace('&', '-')

                        # Create script xml file
                        text = []

                        text.insert(0, f"    <script name='{self.submit_script_name_field.text()}'>")
                        text.insert(1, f"        <script_version>'{self.submit_script_version_lineedit.text()}'</script_version>")
                        text.insert(2, f"        <flame_version>'{self.submit_script_flame_version_lineedit.text()}'</flame_version>")
                        text.insert(3, f"        <date>'{self.submit_script_date_lineedit.text()}'</date>")
                        text.insert(4, f"        <developer>'{self.submit_script_dev_name_lineedit.text()}'</developer>")
                        text.insert(5, f"        <description>'{description_text}'</description>")
                        text.insert(6, '    </script>')

                        out_file = open(script_xml_path, 'w')
                        for line in text:
                            print(line, file=out_file)
                        out_file.close()

                        print('--> Script xml created.\n')

                    def create_tar():

                        pyflame.message_print(
                            message='Creating tar file...',
                            script_name=SCRIPT_NAME,
                            )

                        if self.all_files_btn.isChecked():
                            print('Upload All Files: True\n')

                            # Loop through files in script folder avoiding hidden files and __pycache__ folder
                            tar_file_list = ''
                            skip = ('.', '__pycache__')

                            for file in os.listdir(script_folder):
                                if not any(file.startswith(prefix) for prefix in skip):
                                    tar_file_list += ' ' + file

                            print('tar_file_list:', tar_file_list)
                            tar_file_list.strip()

                            # Create tar command
                            tar_command = f'tar -cvf {script_tar_path} {tar_file_list}'

                        else:
                            print('Upload All Files: False\n')
                            tar_command = 'tar -cvf %s  %s' % (script_tar_path, script_name + '.py')

                        print('Adding files to tar:')

                        os.chdir(script_folder)
                        os.system(tar_command)

                        print('\n')
                        pyflame.message_print(
                            message=f'Python script tar file created.',
                            script_name=SCRIPT_NAME,
                            )

                    def upload_files():

                        pyflame.message_print(
                            message='Uploading script...',
                            script_name=SCRIPT_NAME,
                            )

                        # Connect to ftp
                        self.ftp = FTP(f'{LOGIK_FTP}')
                        self.ftp.login(self.settings.username, self.settings.password)
                        self.ftp.cwd('/Submit_Scripts')

                        pyflame.message_print(
                            message='Uploading Python Script...',
                            script_name=SCRIPT_NAME,
                            )

                        # Close window
                        self.submit_script_window.close()

                        # Upload tgz file to ftp
                        self.upload_file(
                            upload_type='Python Script',
                            file_name=script_name,
                            xml_path=script_xml_path,
                            tgz_path=script_tar_path,
                            )

                        # Bring back main window
                        self.window.show()

                    # Upload script and xml to ftp
                    script_xml_path = os.path.join(self.temp_folder, f'{self.submit_script_path_entry.text().rsplit("/", 1)[1][:-3]}.xml')

                    save_config()

                    create_script_xml(self)

                    script_name = self.submit_script_path_entry.text().rsplit('/', 1)[1][:-3]
                    script_path = self.submit_script_path_entry.text()
                    script_folder = script_path.rsplit('/', 1)[0]

                    script_tar_path = os.path.join(self.temp_folder, script_name) + '.tgz'

                    create_tar()

                    upload_files()

                # Check script path field
                if not os.path.isfile(self.submit_script_path_entry.text()):
                    PyFlameMessageWindow(
                        message='Enter path to python script.',
                        script_name=SCRIPT_NAME,
                        type=MessageType.ERROR,
                        )
                    return

                # Check script version field
                elif not self.submit_script_version_lineedit.text():
                    PyFlameMessageWindow(
                        message='Enter script version.',
                        script_name=SCRIPT_NAME,
                        type=MessageType.ERROR,
                        )
                    return

                # Check script version field for alpha characters
                alpha = [n for n in self.submit_script_version_lineedit.text() if n.isalpha()]
                if alpha:
                    PyFlameMessageWindow(
                        message='Script Version should be numbers only. Such as: 1.0.0',
                        script_name=SCRIPT_NAME,
                        type=MessageType.ERROR,
                        )
                    return

                # Check flame version field
                if not self.submit_script_flame_version_lineedit.text():
                    PyFlameMessageWindow(
                        message='Enter minimum version of Flame needed to run script.',
                        script_name=SCRIPT_NAME,
                        type=MessageType.ERROR,
                        )
                    return

                # Check flame version field for alpha characters
                alpha = [n for n in self.submit_script_version_lineedit.text() if n.isalpha()]
                if alpha:
                    PyFlameMessageWindow(
                        message='Flame Version should be numbers only. Such as: 2021.2.',
                        script_name=SCRIPT_NAME,
                        type=MessageType.ERROR,
                        )
                    return

                # Check script date field
                if not self.submit_script_date_lineedit.text():
                    PyFlameMessageWindow(
                        message='Enter date script was written or updated. Whichever is later.',
                        script_name=SCRIPT_NAME,
                        type=MessageType.ERROR,
                        )
                    return

                # Check date field for proper formatting
                if not re.search('^\\d{2}.\\d{2}.\\d{2}',self.submit_script_date_lineedit.text()):
                    PyFlameMessageWindow(
                        message='Script date should be entered in dd.mm.yy format.',
                        script_name=SCRIPT_NAME,
                        type=MessageType.ERROR,
                        )
                    return

                if not len(self.submit_script_date_lineedit.text()) == 8:
                    PyFlameMessageWindow(
                        message='Script date should be entered in dd.mm.yy format.',
                        script_name=SCRIPT_NAME,
                        type=MessageType.ERROR,
                        )
                    return

                # Check script dev field
                if not self.submit_script_dev_name_lineedit.text():
                    PyFlameMessageWindow(
                        message='Enter name of script author.',
                        script_name=SCRIPT_NAME,
                        type=MessageType.ERROR,
                        )
                    return

                # Check script description field
                elif not self.submit_script_description_text_edit.toPlainText():
                    PyFlameMessageWindow(
                        message='Enter description of script and any notes on working with script.',
                        script_name=SCRIPT_NAME,
                        type=MessageType.ERROR,
                        )
                    return

                # If script already exists on ftp confirm upload
                elif self.submit_script_name_field.text() in self.ftp_script_list:
                    if PyFlameMessageWindow(
                        message='Script already exists on Logik Portal. Update?',
                        script_name=SCRIPT_NAME,
                        type=MessageType.CONFIRM,
                        ):
                        upload()
                else:
                    upload()

            def update_script_info():
                """
                Clean submit window fields and get selected script info.
                """

                # Clear submit fields
                self.submit_script_version_lineedit.setText('')
                self.submit_script_flame_version_lineedit.setText('')
                self.submit_script_date_lineedit.setText('')
                self.submit_script_dev_name_lineedit.setText('')
                self.submit_script_description_text_edit.setPlainText('')

                # Get selected script info
                self.get_script_info()

            def close_submit_window():

                # Close submit window
                self.submit_script_window.close()

                # Bring back main window
                self.window.show()

            # Create submit window
            self.submit_script_window = PyFlameWindow(
                width=800, height=570,
                title='Submit Python Script',
                return_pressed=upload_script
                )

            # Labels
            self.submit_script_label = PyFlameLabel(
                text='Logik Portal Python Script Submit',
                style=Style.UNDERLINE,
                max_width=True,
                )
            self.submit_script_path_label = PyFlameLabel(
                text='Script Path',
                width=110,
                )
            self.submit_script_name_label_01 = PyFlameLabel(
                text='Script Name',
                width=110,
                )
            self.submit_script_version_label_01 = PyFlameLabel(
                text='Script Version',
                width=110,
                )
            self.submit_script_flame_version_label_01 = PyFlameLabel(
                text='Flame Version',
                width=110,
                )
            self.submit_script_date_label_01 = PyFlameLabel(
                text='Date',
                width=110,
                )
            self.submit_script_dev_name_label_01 = PyFlameLabel(
                text='Dev Name',
                width=110,
                )
            self.submit_script_description_label = PyFlameLabel(
                text='Description',
                width=110,
                )

            # Entries
            self.submit_script_name_field = PyFlameLineEdit(
                text='',
                width=400,
                read_only=True,
                )
            self.submit_script_version_lineedit = PyFlameLineEdit(
                text='',
                width=400,
                )
            self.submit_script_flame_version_lineedit = PyFlameLineEdit(
                text='',
                width=400,
                )
            self.submit_script_date_lineedit = PyFlameLineEdit(
                text='',
                width=400,
                )
            self.submit_script_dev_name_lineedit = PyFlameLineEdit(
                text='',
                width=400,
                )

            # Entry File Browser
            self.submit_script_path_entry = PyFlameLineEditFileBrowser(
                text=self.settings.script_submit_path,
                width=400,
                connect=update_script_info,
                browser_type=BrowserType.FILE,
                browser_ext=['py'],
                browser_title='Select Python File',
                browser_window_to_hide=[self.window, self.submit_script_window],
                )

            # TextEdit
            self.submit_script_description_text_edit = PyFlameTextEdit(
                text=self.file_description,
                width=400,
                max_height=True,
                read_only=False,
                )

            # Push Buttons
            self.all_files_btn = PyFlamePushButton(
                text='All Files',
                button_checked=self.settings.python_submit_all_files,
                )

            # Buttons
            self.submit_script_upload_btn = PyFlameButton(
                text='Upload',
                connect=upload_script,
                color=Color.BLUE,
                )
            self.submit_script_cancel_btn = PyFlameButton(
                text='Cancel',
                connect=close_submit_window,
                )

            #------------------------------------#

            # Window Layout
            grid_layout = QtWidgets.QGridLayout()

            grid_layout.addWidget(self.submit_script_label, 0, 0, 1, 6)

            grid_layout.addWidget(self.submit_script_path_label, 1, 0)
            grid_layout.addWidget(self.submit_script_name_label_01, 2, 0)
            grid_layout.addWidget(self.submit_script_version_label_01, 3, 0)
            grid_layout.addWidget(self.submit_script_flame_version_label_01, 4, 0)
            grid_layout.addWidget(self.submit_script_date_label_01, 5, 0)
            grid_layout.addWidget(self.submit_script_dev_name_label_01, 6, 0)
            grid_layout.addWidget(self.submit_script_description_label, 7, 0)

            grid_layout.addWidget(self.submit_script_path_entry, 1, 1, 1, 4)
            grid_layout.addWidget(self.submit_script_name_field, 2, 1, 1, 4)
            grid_layout.addWidget(self.submit_script_version_lineedit, 3, 1, 1, 4)
            grid_layout.addWidget(self.submit_script_flame_version_lineedit, 4, 1, 1, 4)
            grid_layout.addWidget(self.submit_script_date_lineedit, 5, 1, 1, 4)
            grid_layout.addWidget(self.submit_script_dev_name_lineedit, 6, 1, 1, 4)
            grid_layout.addWidget(self.submit_script_description_text_edit, 7, 1, 6, 4)

            grid_layout.addWidget(self.all_files_btn, 2, 5)
            grid_layout.addWidget(self.submit_script_upload_btn, 11, 5)
            grid_layout.addWidget(self.submit_script_cancel_btn, 12, 5)

            # Add layouts to main window
            self.submit_script_window.add_layout(grid_layout)
            self.submit_script_window.show()

            if self.submit_script_path_entry.text().endswith('.py'):
                if os.path.isfile(self.submit_script_path_entry.text()):
                    self.update_script_info()

        def install_local_script():

            def save_config():

                # Save path to config file
                self.settings.save_config(
                    script_name=SCRIPT_NAME,
                    script_path=SCRIPT_PATH,
                    config_values={
                        'script_install_local_path': script_path.rsplit('/', 1)[0]
                        }
                    )

            script_path = pyflame.file_browser(
                title='Select Python File',
                extension=['py'],
                path=self.settings.script_install_local_path,
                window_to_hide=[self.window]
            )

            if script_path:

                save_config()

                script_to_install = script_path.rsplit('/', 1)[1][:-3]
                if PyFlameMessageWindow(
                    message=f'Install Python Script: <b>{script_to_install.replace("_", " ")}',
                    type=MessageType.CONFIRM
                    ):
                    dest_folder = os.path.join(self.shared_script_path, script_to_install)
                    if os.path.isdir(dest_folder):
                        if not PyFlameMessageWindow(
                            message='Python script already exists. Overwrite?',
                            script_name=SCRIPT_NAME,
                            type=MessageType.CONFIRM
                            ):
                            pyflame.message_print(
                                message='Python script not installed.',
                                script_name=SCRIPT_NAME,
                                )
                            return
                        else:
                            shutil.rmtree(dest_folder)

                    # Create local folder for script
                    if not os.path.exists(dest_folder):
                        try:
                            os.makedirs(dest_folder)
                        except:
                            PyFlameMessageWindow(
                                message='Could not create script folder.<br><br>Check path and permissions.',
                                script_name=SCRIPT_NAME,
                                type=MessageType.ERROR,
                            )
                            return

                    # Copy script to dest folder
                    shutil.copy(script_path, dest_folder)

                    # Refresh installed scripts tree list
                    self.get_installed_scripts(self.installed_scripts_tree, self.shared_script_path)

                    # Refresh python hooks
                    flame.execute_shortcut('Rescan Python Hooks')
                    pyflame.message_print(
                        message='Python Hooks Refreshed.',
                        script_name=SCRIPT_NAME,
                        )

                    if os.path.isfile(os.path.join(dest_folder, script_to_install + '.py')):
                        PyFlameMessageWindow(
                            message=f'Python script installed: <b>{script_to_install.replace("_", " ")}',
                            script_name=SCRIPT_NAME,
                            type=MessageType.INFO,
                        )
                        return
                    PyFlameMessageWindow(
                        message=f'Python script not installed.',
                        script_name=SCRIPT_NAME,
                        type=MessageType.ERROR,
                    )
                    return

        def download_script(portal_tree, installed_tree, script_path):

            # Get selected script info from selection
            selected_script = portal_tree.selectedItems()
            script_item = selected_script[0]
            script_name = script_item.text(0).strip()
            script_name = script_name.replace(' ', '_')
            script_flame_version = script_item.text(2)

            print('Script name:', script_name)
            print('Script Flame version:', script_flame_version, '\n')

            # Set source and destination paths
            source_script_path = os.path.join('/Scripts', script_flame_version, script_name) + '.tgz'
            #print('source_script_path: ', source_script_path)

            dest_script_path = os.path.join(script_path, script_name, script_name) + '.tgz'
            #print('dest_script_path: ', dest_script_path)

            dest_folder = dest_script_path.rsplit('/', 1)[0]
            #print('dest_folder:', dest_folder)

            if os.path.isdir(dest_folder):
                if not PyFlameMessageWindow(
                    message='Script already exists. Overwrite?',
                    script_name=SCRIPT_NAME,
                    type=MessageType.CONFIRM,
                    ):
                    pyflame.message_print(
                        message='Script not installed.',
                        script_name=SCRIPT_NAME,
                        )
                    return
                else:
                    shutil.rmtree(dest_folder)

            # Create local folder for script
            try:
                os.makedirs(dest_folder)
            except:
                pass

            pyflame.message_print(
                message=f'Downloading python script...',
                script_name=SCRIPT_NAME,
                )

            # Download python script tgz file from ftp
            self.download_file(
                download_type='Python Script',
                file_name=script_name,
                ftp_file_path=source_script_path,
                tgz_path=dest_script_path,
                )

            # Uncompress tgz file
            tar_command = f'tar -xf {dest_script_path} -C {dest_folder}'
            os.system(tar_command)

            # delete tgz file
            os.remove(dest_script_path)

            print('--> Script installed.\n')

            # Refresh installed scripts tree list

            self.get_installed_scripts(installed_tree, script_path)

            # Set color of selected script in portal tree to normal color

            script_item.setForeground(0, QtGui.QColor(154, 154, 154))

            # Refresh python hooks

            flame.execute_shortcut('Rescan Python Hooks')

            pyflame.message_print(
                message='Python hooks refreshed.',
                script_name=SCRIPT_NAME,
                )

            if os.path.isfile(dest_script_path[:-3] + 'py'):
                return
            PyFlameMessageWindow(
                message='Download failed.',
                script_name=SCRIPT_NAME,
                type=MessageType.ERROR,
            )
            return

        # Tab 1

        # Labels
        self.installed_scripts_label = PyFlameLabel(
            text='Installed Python Scripts',
            style=Style.UNDERLINE,
            width=525,
            )
        self.portal_scripts_label = PyFlameLabel(
            text='Portal Python Scripts',
            style=Style.UNDERLINE,
            width=525,
            )
        self.script_description_label = PyFlameLabel(
            text='Python Script Description',
            style=Style.UNDERLINE,
            max_width=True,
            )

        # Logo
        self.logik_logo_label = QtWidgets.QLabel()
        self.logo_label_pixmap = QtGui.QPixmap(os.path.join(SCRIPT_PATH, 'logik_logo.png'))
        self.logik_logo_label.setPixmap(self.logo_label_pixmap)

        # Text Edit
        self.script_description_text_edit = PyFlameTextEdit(
            text=self.file_description,
            height=200,
            max_width=True,
            read_only=True,
            )

        # Installed Scripts TreeWidget
        self.installed_scripts_tree = PyFlameTreeWidget(
            column_names=[
            'Name',
            'Version',
            'Flame',
            'Date',
            'Author',
            'Path',
            ],
            connect=self.installed_script_description,
            width=525,
            height=250,
            )

        # Available Scripts TreeWidget
        self.portal_scripts_tree = PyFlameTreeWidget(
            column_names=[
            'Name',
            'Version',
            'Flame Min',
            'Flame Max',
            'Date',
            'Author',
            ],
            connect=self.get_script_description,
            width=525,
            height=250,
            )

        # Buttons
        self.script_submit_btn = PyFlameButton(
            text='Submit',
            connect=login_check,
            )
        self.install_script_btn = PyFlameButton(
            text='Install',
            connect=partial(
                download_script,
                self.portal_scripts_tree,
                self.installed_scripts_tree,
                self.shared_script_path
                ),
            color=Color.BLUE,
            )
        self.install_local_script_btn = PyFlameButton(
            text='Install Local',
            connect=install_local_script,
            )
        self.delete_script_btn = PyFlameButton(
            text='Delete',
            connect=partial(
                self.delete_script,
                self.installed_scripts_tree,
                self.shared_script_path
                ),
            )
        self.python_done_btn = PyFlameButton(
            text='Done',
            connect=self.done,
            )

        # Disable script download button if current flame version older than script minimum
        self.portal_scripts_tree.clicked.connect(
            partial(
                self.check_script_flame_version,
                self.portal_scripts_tree
                )
            )

        #  UI Widget Layout
        self.window.tab1.layout = QtWidgets.QGridLayout()

        self.window.tab1.layout.addWidget(self.installed_scripts_label, 0, 0, 1, 5)
        self.window.tab1.layout.addWidget(self.portal_scripts_label, 0, 7, 1, 5)

        self.window.tab1.layout.addWidget(self.installed_scripts_tree, 1, 0, 1, 5)
        self.window.tab1.layout.addWidget(self.portal_scripts_tree, 1, 7, 1, 5)

        self.window.tab1.layout.addWidget(self.delete_script_btn, 2, 0)
        self.window.tab1.layout.addWidget(self.install_local_script_btn, 2, 4)

        self.window.tab1.layout.addWidget(self.script_submit_btn, 2, 7)
        self.window.tab1.layout.addWidget(self.install_script_btn, 2, 11)

        self.window.tab1.layout.addWidget(self.script_description_label, 4, 0, 1, 12)
        self.window.tab1.layout.addWidget(self.script_description_text_edit, 5, 0, 1, 12)

        self.window.tab1.layout.addWidget(self.logik_logo_label, 6, 0)
        self.window.tab1.layout.addWidget(self.python_done_btn, 6, 11)

        self.window.tab1.setLayout(self.window.tab1.layout)

    def matchbox_tab(self):

        def download_logik_collection():

            def save_config():

                # Save path to config file
                self.settings.save_config(
                    script_name=SCRIPT_NAME,
                    script_path=SCRIPT_PATH,
                    config_values={
                        'matchbox_path': self.matchbox_install_path
                        }
                    )

            def download():

                # Change cursor to busy
                QtWidgets.QApplication.setOverrideCursor(QtCore.Qt.BusyCursor)

                pyflame.message_print(
                    message='Downloading Logik Matchboxes from logik-matchbook.org...',
                    script_name=SCRIPT_NAME,
                    )

                # Path to download from ftp
                ftp_file_path = os.path.join('/Logik_Matchbox/MatchboxShaderCollection.tgz')

                # Download dest path
                tgz_path = '/opt/Autodesk/shared/python/logik_portal/MatchboxShaderCollection.tgz'

                # Download batch tgz file from ftp
                self.download_file(
                    download_type='Matchbox',
                    file_name='Logik Matchbox',
                    ftp_file_path=ftp_file_path,
                    tgz_path=tgz_path,
                    )

                # Untar matchbox archive
                install_path = os.path.join(self.matchbox_install_path, 'LOGIK')

                command = f'tar -xvpzf /opt/Autodesk/shared/python/logik_portal/MatchboxShaderCollection.tgz --strip-components 1 -C {install_path}'
                command = command.split(' ', 6)

                create_dir_command = f'mkdir -p {install_path}'
                create_dir_command = create_dir_command.split(' ', 3)

                if not folder_write_permission:
                    try:
                        # use popen to run create_dir_command as sudo
                        p = Popen(['sudo', '-S'] + create_dir_command, stdin=PIPE, stderr=PIPE, universal_newlines=True)
                        p.communicate(system_password + '\n')[1]

                        # Sudo untar
                        p = Popen(['sudo', '-S'] + command, stdin=PIPE, stderr=PIPE, universal_newlines=True)
                        p.communicate(system_password + '\n')[1]

                        QtWidgets.QApplication.restoreOverrideCursor()

                    except:
                        QtWidgets.QApplication.restoreOverrideCursor()
                        PyFlameMessageWindow(
                            message=f'Logik Matchbox Collection<br><br>{install_path}',
                            title=f'{SCRIPT_NAME}: Install Failed',
                            type=MessageType.ERROR
                            )
                else:
                    try:
                        # Create install_path directory if it doesn't exist
                        if not os.path.isdir(install_path):
                            os.makedirs(install_path)

                        # Normal untar
                        Popen(command, stdin=PIPE, stderr=PIPE, universal_newlines=True)
                        QtWidgets.QApplication.restoreOverrideCursor()
                    except:
                        QtWidgets.QApplication.restoreOverrideCursor()
                        PyFlameMessageWindow(
                            message=f'Logik Matchbox Collection<br><br>{install_path}',
                            title=f'{SCRIPT_NAME}: Install Failed',
                            type=MessageType.ERROR
                            )
                #self.progress_window.enable_done_button(True)
                os.remove(tgz_path)

            path = pyflame.file_browser(
                title='Select Logik Matchbox Install Directory',
                path=self.settings.matchbox_path,
                select_directory=True,
                window_to_hide=[self.window]
                )
            if path:
                self.matchbox_install_path = path
            else:
                return

            save_config()

            folder_write_permission = os.access(self.matchbox_install_path, os.W_OK)

            if not folder_write_permission:
                matchbox_password_window = PyFlamePasswordWindow(message=f'System password needed to install Logik Matchboxes to selected location.')
                system_password = matchbox_password_window.password()
                if system_password:
                    download()
            else:
                download()

        def add_matchbox_to_batch():

            def create_matchbox_node(matchbox_file_name):

                # Get cursor position

                cursor_pos = flame.batch.cursor_position

                # Create matchbox node

                matchbox_node = flame.batch.create_node('Matchbox', os.path.join(temp_matchbox_path, matchbox_file_name))
                matchbox_node.pos_x = cursor_pos[0]
                matchbox_node.pos_y = cursor_pos[1]

                matchbox_node.load_node_setup(os.path.join(temp_matchbox_path, selected_matchbox_name))

            # Switch to batch tab
            flame.set_current_tab('Batch')

            # Get selected matchbox name
            selected_matchbox_name = self.matchbox_tree.selectedItems()[0].text(0)

            # Connect to ftp
            self.ftp_download_connect()

            # Get list of matchbox files from ftp
            matchbox_files = self.ftp.nlst('/Logik_Matchbox/Shaders/')
            #print('matchbox_files:', matchbox_files, '\n')

            # Get selected matchbox files
            selected_matchbox_files = [file for file in matchbox_files if selected_matchbox_name in file]
            #print('selected_matchbox_files:', selected_matchbox_files, '\n')

            # Create temp matchbox directory in temp folder
            temp_matchbox_path = os.path.join(self.temp_folder, selected_matchbox_name)
            if not os.path.isdir(temp_matchbox_path):
                os.makedirs(temp_matchbox_path)

            # Download selected matchbox files from ftp to temp matchbox directory
            for file_path in selected_matchbox_files:
                file_name = file_path.split('/')[-1]
                with open(os.path.join(temp_matchbox_path, file_name), 'wb') as local_file:
                    self.ftp.retrbinary("RETR " + file_path, local_file.write)

            # Get name of glsl file or mx file to load into matchbox node
            glsl_files = [file for file in selected_matchbox_files if file.endswith('.glsl')]

            if glsl_files != []:
                glsl_file = glsl_files[-1]
                matchbox_file_name = glsl_file.split('/')[-1]
            else:
                matchbox_file_name = [file for file in selected_matchbox_files if file.endswith('.mx')][0].rsplit('/', 1)[-1]
            #print('matchbox_file_name:', matchbox_file_name, '\n')

            create_matchbox_node(matchbox_file_name)
            PyFlameMessageWindow(
                message=f'''Matchbox added to Batch: <b>{selected_matchbox_name}</b><br><br>
                            You may have to click in Batch for the node to appear.''',
                script_name=SCRIPT_NAME,
                )

        def matchbox_search():

            self.update_matchbox_tree(search=self.matchbox_search_entry.text())

        # Tab 3

        # Labels
        self.matchbox_search_label = PyFlameLabel(
            text='Search',
            width=75,
            )
        self.matchbox_logik_matchbox_collection_label = PyFlameLabel(
            text='Logik Matchbox Collection',
            style=Style.UNDERLINE,
            max_width=True,
            )
        self.matchbox_desciption_label = PyFlameLabel(
            text='Matchbox Description',
            style=Style.UNDERLINE,
            max_width=True,
            )

        # Logik Logo
        self.logik_logo_label = QtWidgets.QLabel()
        self.logo_label_pixmap = QtGui.QPixmap(os.path.join(SCRIPT_PATH, 'logik_logo.png'))
        self.logik_logo_label.setPixmap(self.logo_label_pixmap)

        # Entry
        self.matchbox_search_entry = PyFlameLineEdit(
            text='',
            max_width=True,
            text_changed=matchbox_search,
            )

        # Text Edit
        self.matchbox_text_edit = PyFlameTextEdit(
            text=self.file_description,
            height=200,
            max_width=True,
            read_only=True,
            )

        # Matchbox TreeWidget
        logik_matchbox_columns = [
            'Matchbox Name',
            'Shader Type',
            'Author',
            ]
        self.matchbox_tree = PyFlameTreeWidget(
            column_names=logik_matchbox_columns,
            connect=self.get_matchbox_description,
            height=250,
            max_width=True,
            )
        self.matchbox_tree.setColumnWidth(0, 400)
        self.matchbox_tree.setColumnWidth(1, 300)
        self.matchbox_tree.setColumnWidth(2, 400)
        self.matchbox_tree.setTextElideMode(QtCore.Qt.ElideNone)

        # Buttons
        self.matchbox_add_to_batch = PyFlameButton(
            text='Add to Batch',
            connect=add_matchbox_to_batch,
            )
        self.matchbox_download_all_btn = PyFlameButton(
            text='Download All',
            connect=download_logik_collection,
            color=Color.BLUE,
            )
        self.matchbox_done_btn = PyFlameButton(
            text='Done',
            connect=self.done,
            )

        # Layout
        search_hbox = QtWidgets.QHBoxLayout()
        search_hbox.addWidget(self.matchbox_search_label)
        search_hbox.addWidget(self.matchbox_search_entry)

        self.window.tab3.layout = QtWidgets.QGridLayout()

        self.window.tab3.layout.addWidget(self.matchbox_logik_matchbox_collection_label, 0, 0, 1, 12)
        self.window.tab3.layout.addWidget(self.matchbox_tree, 1, 0, 1, 12)

        self.window.tab3.layout.addLayout(search_hbox, 2, 0)
        self.window.tab3.layout.addWidget(self.matchbox_add_to_batch, 2, 10)
        self.window.tab3.layout.addWidget(self.matchbox_download_all_btn, 2, 11)

        self.window.tab3.layout.addWidget(self.matchbox_desciption_label, 4, 0, 1, 12)
        self.window.tab3.layout.addWidget(self.matchbox_text_edit, 5, 0, 1, 12)

        self.window.tab3.layout.addWidget(self.logik_logo_label, 6, 0)
        self.window.tab3.layout.addWidget(self.matchbox_done_btn, 6, 11)

        self.window.tab3.setLayout(self.window.tab3.layout)

    def batch_setups_tab(self):

        def batch_setups_download():

            def save_config():

                # Save config
                self.settings.save_config(
                    script_name=SCRIPT_NAME,
                    script_path=SCRIPT_PATH,
                    config_values={
                        'batch_setup_download_path': download_path,
                        'open_batch': str(self.open_batch_btn.isChecked())
                        }
                    )

            def open_batch():

                # Get batch setup path
                for f in os.listdir(self.settings.batch_setup_download_path):
                    if f.split('.', 1)[0] == batch_name and f.endswith('.batch'):
                        setup_path = os.path.join(self.settings.batch_setup_download_path, f)
                        # print('setup_path:', setup_path)

                # Create new batch group
                # Names for shelf and schematic reels can be added or deleted here
                # Each reel name must be in quotes and seperated by commas
                schematic_reel_list = ['Plates', 'Elements', 'PreRenders', 'Ref']
                shelf_reel_list = ['Batch Renders']

                self.batch_group = flame.batch.create_batch_group(str(batch_item.text(0)), duration=100, reels=schematic_reel_list, shelf_reels=shelf_reel_list)

                # Load batch setup
                self.batch_group.load_setup(setup_path)

                pyflame.message_print(
                    message='Batch setup loaded.',
                    script_name=SCRIPT_NAME,
                    )

            download_path = pyflame.file_browser(
                title='Batch Setup Download Path',
                path=self.settings.batch_setup_download_path,
                select_directory=True,
                window_to_hide=[self.window]
                )

            if download_path:

                save_config()

                # Get batch info from script description
                selected_batch = self.batch_setups_tree.selectedItems()
                batch_item = selected_batch[0]
                batch_name = batch_item.text(0)

                # Check path to see if batch already exists
                batch_exists = [b for b in os.listdir(self.settings.batch_setup_download_path) if b.split('.', 1)[0] == batch_name]

                # If batch already exists prompt to overwrite or cancel
                if batch_exists:
                    if not PyFlameMessageWindow(
                        message='Batch Already Exists. Overwrite?',
                        script_name=SCRIPT_NAME,
                        type=MessageType.CONFIRM
                        ):
                        pyflame.message_print(
                            message='Batch setup download cancelled.',
                            script_name=SCRIPT_NAME,
                            )
                        return
                    else:
                        for b in batch_exists:
                            path_to_delete = os.path.join(self.settings.batch_setup_download_path, b)
                            if os.path.isfile(path_to_delete):
                                os.remove(path_to_delete)
                            elif os.path.isdir(path_to_delete):
                                shutil.rmtree(path_to_delete)

                pyflame.message_print(
                    message=f'Downloading batch setup...',
                    script_name=SCRIPT_NAME,
                    )

                # Path to download from ftp
                ftp_file_path = os.path.join('/Batch_Setups', batch_item.text(1), batch_name + '.tgz')

                # Download dest path
                tgz_path = os.path.join(self.settings.batch_setup_download_path, batch_name + '.tgz')

                # Download batch tgz file from ftp
                self.download_file(
                    download_type='Batch Setup',
                    file_name=batch_name,
                    ftp_file_path=ftp_file_path,
                    tgz_path=tgz_path,
                    )

                # Uncompress tgz file
                tgz_escaped_path = tgz_path.replace(' ', '\ ')
                download_escaped_path = self.settings.batch_setup_download_path.replace(' ', '\ ') + '/'
                tar_command = f'tar -xvf {tgz_escaped_path} -C {download_escaped_path}'
                os.system(tar_command)

                # Delete tgz file
                os.remove(tgz_path)

                if self.open_batch_btn.isChecked():
                    open_batch()

        def submit_batch_setup():

            def get_batch_name():
                """
                Fill in Batch Name field once valid batch path is entered. Must end with .batch
                If .flare is in file path, remove it
                """

                if self.submit_batch_path_entry.text().endswith('.batch'):
                    batch_name = self.submit_batch_path_entry.text().rsplit('/', 1)[1][:-6]
                    if batch_name.endswith('.flare'):
                        batch_name = batch_name[:-6]
                    self.submit_batch_name_field.setText(batch_name)
                else:
                    self.submit_batch_name_field.setText('')

            def get_batch_setup_info():
                """
                Get selected batch setup info.
                """

                batch_setup_path = self.submit_batch_path_entry.text()

                if batch_setup_path:
                    self.submit_batch_path_entry.setText(batch_setup_path)
                    batch_name = batch_setup_path.rsplit('/', 1)[1][:-6]
                    if batch_name.endswith('.flare'):
                        batch_name = batch_name[:-6]
                    self.submit_batch_name_field.setText(batch_name)

            def batch_setup_upload():

                def save_config():

                    # Save path to config file
                    self.settings.save_config(
                        script_name=SCRIPT_NAME,
                        script_path=SCRIPT_PATH,
                        config_values={
                            'batch_submit_path': self.submit_batch_path_entry.text()
                            }
                        )

                def compress_batch_setup():

                    # Add batch files to tar file

                    batch_folder_path = self.submit_batch_path_entry.text()[:-6]
                    print('Batch folder path:', batch_folder_path)

                    batch_root_folder = batch_folder_path.rsplit('/', 1)[0]
                    batch_folder = batch_folder_path.rsplit('/', 1)[1]

                    tar_file_list = batch_folder + ' ' + batch_folder + '.batch'
                    print('Tar file list:', tar_file_list)

                    if batch_folder_path.endswith('.flare'):
                        self.tar_file_name = batch_folder_path.rsplit('/', 1)[1][:-6]
                    else:
                        self.tar_file_name = batch_folder_path.rsplit('/', 1)[1]
                    print('Tar file name:', self.tar_file_name)

                    self.tar_path = os.path.join(self.temp_folder, self.tar_file_name) + '.tgz'
                    print('Tar path:', self.tar_path)

                    tar_command = f'tar -cvf {self.tar_path}  {tar_file_list}'
                    print('Tar command:', tar_command)

                    os.chdir(batch_root_folder)
                    os.system(tar_command)

                    print('\n')
                    pyflame.message_print(
                        message='Batch tar file created.',
                        script_name=SCRIPT_NAME,
                        )

                def create_batch_xml_file():

                    description_text = self.submit_batch_description_text_edit.toPlainText()
                    description_text = description_text.replace("'", "\"")
                    description_text = description_text.replace('&', '-')

                    # Create batch info file

                    text = []

                    text.insert(0, f"    <batch name='{self.tar_file_name}'>")
                    text.insert(1, f"        <artist>'{self.submit_batch_artist_name_entry.text()}'</artist>")
                    text.insert(2, f"        <flame_version>'{self.submit_batch_flame_version_field.text()}'</flame_version>")
                    text.insert(3, f"        <description>'{description_text}'</description>")
                    text.insert(4, '    </batch>')

                    out_file = open(xml_file, 'w')
                    for line in text:
                        print(line, file=out_file)
                    out_file.close()

                def upload_batch():

                    self.ftp_upload_connect()

                    # Check to see if file already exists on ftp
                    ftp_file_list = self.ftp.nlst('/Submit_Batch_Setups')

                    if self.tar_file_name + '.tgz' in ftp_file_list:
                        QtWidgets.QApplication.restoreOverrideCursor()
                        self.ftp.quit()
                        PyFlameMessageWindow(
                            message='Batch setup already exists. Rename and try again.',
                            script_name=SCRIPT_NAME,
                            type=MessageType.ERROR
                            )
                        return

                    pyflame.message_print(
                        message='Uploading Batch Setup...',
                        script_name=SCRIPT_NAME,
                        )

                    # Close window
                    self.submit_batch_window.close()

                    # Upload tgz file to ftp
                    self.upload_file(
                        upload_type='Batch Setup',
                        file_name=self.tar_file_name,
                        xml_path=xml_file,
                        tgz_path=self.tar_path,
                        )

                if not os.path.isfile(self.submit_batch_path_entry.text()):
                    PyFlameMessageWindow(
                        message='Enter valid batch setup path.',
                        script_name=SCRIPT_NAME,
                        type=MessageType.ERROR
                        )
                    return
                elif self.submit_batch_artist_name_entry.text() == '':
                    PyFlameMessageWindow(
                        message='Enter Artist name.',
                        script_name=SCRIPT_NAME,
                        type=MessageType.ERROR
                        )
                    return
                elif self.submit_batch_description_text_edit.toPlainText() == '':
                    PyFlameMessageWindow(
                        message='Enter batch setup description.',
                        script_name=SCRIPT_NAME,
                        type=MessageType.ERROR
                        )
                    return
                else:
                    save_config()
                    compress_batch_setup()
                    xml_file = os.path.join(SCRIPT_PATH, f'{self.tar_file_name}.xml')
                    create_batch_xml_file()
                    upload_batch()

            def close_submit_batch_window():

                self.submit_batch_window.close()

                self.window.show()

            flame_version = str(self.flame_version)
            if flame_version.endswith('.0'):
                flame_version = flame_version[:-2]

            # Create Batch Setup Submit Window
            self.submit_batch_window = PyFlameWindow(
                width=800,
                height=530,
                title='Submit Batch Setup',
                return_pressed=batch_setup_upload
                )

            # Labels
            self.submit_batch_label = PyFlameLabel(
                text='Logik Portal Batch Setup Submit',
                style=Style.UNDERLINE,
                max_width=True,
                )
            self.submit_batch_path_label = PyFlameLabel(
                text='Batch Path',
                width=110,
                )
            self.submit_batch_name_label = PyFlameLabel(
                text='Batch Name',
                width=110,
                )
            self.submit_batch_flame_version_label = PyFlameLabel(
                text='Flame Version',
                width=110,
                )
            self.submit_batch_artist_name_label = PyFlameLabel(
                text='Artist Name',
                width=110,
                )
            self.submit_batch_description_label = PyFlameLabel(
                text='Batch Description',
                width=110,
                )

            # Entry Fields
            self.submit_batch_name_field = PyFlameLineEdit(
                text='',
                max_width=True,
                read_only=True,
                )
            self.submit_batch_artist_name_entry = PyFlameLineEdit(
                text='',
                max_width=True,
                )
            self.submit_batch_flame_version_field = PyFlameLineEdit(
                text=flame_version,
                max_width=True,
                read_only=True,
                )

            # Line Edit File Browser
            self.submit_batch_path_entry = PyFlameLineEditFileBrowser(
                text=self.settings.archive_submit_path,
                max_width=True,
                connect=get_batch_setup_info,
                browser_type=BrowserType.FILE,
                browser_ext=['batch'],
                browser_title='Select Batch Setup',
                browser_window_to_hide=[self.window, self.submit_batch_window]
                )

            get_batch_name()

            # TextEdit
            self.submit_batch_description_text_edit = PyFlameTextEdit(
                text=self.file_description,
                max_width=True,
                max_height=True,
                read_only=False,
                )

            # Buttons
            self.submit_batch_upload_btn = PyFlameButton(
                text='Upload',
                connect=batch_setup_upload,
                color=Color.BLUE,
                )
            self.submit_cancel_btn = PyFlameButton(
                text='Cancel',
                connect=close_submit_batch_window,
                )

            # Window Layout
            grid_layout = QtWidgets.QGridLayout()
            grid_layout.setColumnMinimumWidth(2, 150)
            grid_layout.setColumnMinimumWidth(3, 150)
            grid_layout.setRowMinimumHeight(11, 30)

            grid_layout.addWidget(self.submit_batch_label, 0, 0, 1, 6)
            grid_layout.addWidget(self.submit_batch_path_label, 1, 0)
            grid_layout.addWidget(self.submit_batch_name_label, 2, 0)
            grid_layout.addWidget(self.submit_batch_artist_name_label, 3, 0)
            grid_layout.addWidget(self.submit_batch_flame_version_label, 4, 0)
            grid_layout.addWidget(self.submit_batch_description_label, 5, 0)

            grid_layout.addWidget(self.submit_batch_path_entry, 1, 1, 1, 5)
            grid_layout.addWidget(self.submit_batch_name_field, 2, 1, 1, 5)
            grid_layout.addWidget(self.submit_batch_artist_name_entry, 3, 1, 1, 5)
            grid_layout.addWidget(self.submit_batch_flame_version_field, 4, 1, 1, 5)
            grid_layout.addWidget(self.submit_batch_description_text_edit, 5, 1, 6, 5)

            grid_layout.addWidget(self.submit_cancel_btn, 12, 4)
            grid_layout.addWidget(self.submit_batch_upload_btn, 12, 5)

            # Add layout to main window
            self.submit_batch_window.add_layout(grid_layout)
            self.submit_batch_window.show()

        #  Batch Tab 4

        # Labels
        self.batch_setups_label = PyFlameLabel(
            text='Batch Setups',
            style=Style.UNDERLINE,
            max_width=True,
            )
        self.batch_setups_desciption_label = PyFlameLabel(
            text='Batch Setup Description',
            style=Style.UNDERLINE,
            max_width=True,
            )

        # Logo
        self.logik_logo_label = QtWidgets.QLabel()
        self.logo_label_pixmap = QtGui.QPixmap(os.path.join(SCRIPT_PATH, 'logik_logo.png'))
        self.logik_logo_label.setPixmap(self.logo_label_pixmap)

        # Text Edit
        self.batch_setups_text_edit = PyFlameTextEdit(
            text=self.file_description,
            height=200,
            max_width=True,
            read_only=True,
            )

        # Batch Setups TreeWidget
        batch_setup_columns = [
            'Name',
            'Flame',
            'Artist',
            ]
        self.batch_setups_tree = PyFlameTreeWidget(
            column_names=batch_setup_columns,
            connect=self.get_batch_description,
            height=250,
            max_width=True,
            )

        self.batch_setups_tree.setColumnWidth(0, 600)
        self.batch_setups_tree.setColumnWidth(1, 100)
        self.batch_setups_tree.setColumnWidth(2, 300)
        self.batch_setups_tree.setTextElideMode(QtCore.Qt.ElideNone)

        # Disable batch download button if current flame version older than batch minimum
        self.batch_setups_tree.clicked.connect(
            partial(
                self.check_batch_flame_version,
                self.batch_setups_tree
                )
            )

        # Push Buttons
        self.open_batch_btn = PyFlamePushButton(
            text=' Open Batch',
            button_checked=self.settings.open_batch,
            tooltip='Opens batch setup after download is finished',
            )

        # Buttons
        self.batch_setups_submit_btn = PyFlameButton(
            text='Submit',
            connect=submit_batch_setup,
            )
        self.batch_setups_download_btn = PyFlameButton(
            text='Download',
            connect=batch_setups_download,
            color=Color.BLUE,
            )
        self.batch_done_btn = PyFlameButton(
            text='Done',
            connect=self.done,
            )

        #------------------------------------#

        #  Tab layout

        self.window.tab4.layout = QtWidgets.QGridLayout()

        self.window.tab4.layout.addWidget(self.batch_setups_label, 0, 0, 1, 5)

        self.window.tab4.layout.addWidget(self.batch_setups_tree, 1, 0, 1, 5)

        self.window.tab4.layout.addWidget(self.batch_setups_submit_btn, 2, 0)

        self.window.tab4.layout.addWidget(self.open_batch_btn, 2, 3)
        self.window.tab4.layout.addWidget(self.batch_setups_download_btn, 2, 4)

        self.window.tab4.layout.addWidget(self.batch_setups_desciption_label, 4, 0, 1, 5)
        self.window.tab4.layout.addWidget(self.batch_setups_text_edit, 5, 0, 1, 5)

        self.window.tab4.layout.addWidget(self.logik_logo_label, 6, 0)
        self.window.tab4.layout.addWidget(self.batch_done_btn, 6, 4)

        self.window.tab4.setLayout(self.window.tab4.layout)

    def archives_tab(self):

        def archive_download():

            def save_config():

                # Save path to config file

                self.settings.save_config(
                    script_name=SCRIPT_NAME,
                    script_path=SCRIPT_PATH,
                    config_values={
                            'archive_download_path': archive_download_path
                            }
                        )

            # Get path to download archive to
            archive_download_path = pyflame.file_browser(
                title='Archive Download Path',
                path=self.settings.archive_download_path,
                select_directory=True,
                window_to_hide=[self.window]
                )

            if archive_download_path:

                save_config()

                # Read description from script
                selected_archive = self.archives_tree.selectedItems()
                archive_item = selected_archive[0]
                archive_name = archive_item.text(0)
                # print('archive_name:', archive_name)

                # Check path to see if batch already exists
                archive_exists = [a for a in os.listdir(archive_download_path) if a.split('.', 1)[0] == archive_name]
                # print('archive_exists:', archive_exists)

                # If archive already exists prompt to overwrite or cancel
                if archive_exists:
                    if not PyFlameMessageWindow(
                        message='Archive Already Exists. Overwrite?',
                        script_name=SCRIPT_NAME,
                        type=MessageType.CONFIRM
                        ):
                        pyflame.message_print(
                            message='Archive download cancelled.',
                            script_name=SCRIPT_NAME,
                            )
                        return
                    else:
                        for a in archive_exists:
                            path_to_delete = os.path.join(archive_download_path, a)
                            if os.path.isfile(path_to_delete):
                                os.remove(path_to_delete)
                            elif os.path.isdir(path_to_delete):
                                shutil.rmtree(path_to_delete)
                        print('\n')

                pyflame.message_print(
                    message=f'Downloading archive...',
                    script_name=SCRIPT_NAME,
                    )

                # Path to download
                ftp_file_path = os.path.join('/Archives', archive_item.text(1), archive_name + '.tgz')

                # Download dest path
                tgz_path = os.path.join(archive_download_path, archive_name + '.tgz')

                # Download archive tgz file
                self.download_file(
                    download_type='Archive',
                    file_name=archive_name,
                    ftp_file_path=ftp_file_path,
                    tgz_path=tgz_path,
                    )

                # Uncompress tgz file
                tgz_escaped_path = tgz_path.replace(' ', '\ ')
                download_escaped_path = archive_download_path.replace(' ', '\ ') + '/'
                tar_command = f'tar -xvf {tgz_escaped_path} -C {download_escaped_path}'
                os.system(tar_command)

                # Delete tgz file
                os.remove(tgz_path)

        def submit_archive():

            def get_archive_name():

                archive_path = self.submit_archive_path_entry.text()

                if os.path.isdir(archive_path):
                    seg_file = [f for f in os.listdir(archive_path) if f.endswith('.seg')]
                    if not seg_file:
                        self.submit_archive_path_entry.setText('')
                        self.submit_archive_name_field.setText('')
                    else:
                        for f in os.listdir(self.submit_archive_path_entry.text()):
                            if not f.startswith('.'):
                                file_ext = os.path.splitext(f)[-1]
                                if '' == file_ext:
                                    self.submit_archive_name_field.setText(f)
                else:
                    self.submit_archive_path_entry.setText('')
                    self.submit_archive_name_field.setText('')

            def get_archive_info():
                """
                Get info for selected archive.
                """

                # Check selected folder for archive seg file. Give error if not found.
                archive_path = self.submit_archive_path_entry.text()

                if archive_path:
                    seg_file = [f for f in os.listdir(archive_path) if f.endswith('.seg')]
                    if not seg_file:
                        self.submit_archive_path_entry.setText(self.settings.archive_submit_path)
                        self.submit_archive_name_field.setText('')
                        PyFlameMessageWindow(
                            message='Archive not found in selected folder.',
                            script_name=SCRIPT_NAME,
                            type=MessageType.ERROR
                        )
                        return
                    else:
                        self.submit_archive_path_entry.setText(archive_path)
                        for f in os.listdir(archive_path):
                            file_ext = os.path.splitext(f)[-1]
                            if '' == file_ext:
                                self.submit_archive_name_field.setText(f)

            def archive_upload():

                def save_config():

                    # Save path to config file
                    self.settings.save_config(
                        script_name=SCRIPT_NAME,
                        script_path=SCRIPT_PATH,
                        config_values={
                            'archive_submit_path': self.submit_archive_path_entry.text()
                            }
                        )

                def compress_archive():
                    """
                    Create tar file from archive folder
                    """

                    archive_folder = self.submit_archive_path_entry.text()
                    print('Archive folder:', archive_folder, '\n')

                    self.tar_file_name = self.submit_archive_name_field.text()
                    # print('tar_file_name:', self.tar_file_name)

                    self.tar_path = os.path.join(self.temp_folder, self.tar_file_name) + '.tgz'
                    # print('tar_path:', self.tar_path)

                    tar_file_list = ''

                    for f in os.listdir(archive_folder):
                        if not f.startswith('.'):
                            tar_file_list = tar_file_list + ' ' + f
                    tar_file_list.strip()
                    # print('tar_file_list:', tar_file_list)

                    tar_command = f'tar -cvf {self.tar_path}  {tar_file_list}'
                    # print('tar_command:', tar_command)

                    os.chdir(archive_folder)
                    os.system(tar_command)

                    print('\n')
                    pyflame.message_print(
                        message='Tar archive created.',
                        script_name=SCRIPT_NAME,
                        )

                def create_archive_xml_file():

                    description_text = self.submit_archive_description_text_edit.text()
                    description_text = description_text.replace("'", "\"")
                    description_text = description_text.replace('&', '-')

                    # Get tar file size. If file is larger than 2GB, give error

                    tar_file_size = os.path.getsize(self.tar_path)

                    if len(str(tar_file_size)) > 6:
                        tar_file_size = int(str(tar_file_size)[:-6])
                        if tar_file_size > 2000:
                            PyFlameMessageWindow(
                                message='Archive to large. Reduce size and try again.',
                                script_name=SCRIPT_NAME,
                                type=MessageType.ERROR
                            )
                            return
                        else:
                            tar_file_size = str(tar_file_size) + 'mb'
                    else:
                        tar_file_size = '> 1mb'
                    # print('tar_file_size:', tar_file_size)

                    # Create archive info file
                    text = []

                    text.insert(0, f"    <archive name='{self.tar_file_name}'>")
                    text.insert(1, f"        <flame_version>'{self.submit_archive_flame_version_field.text()}'</flame_version>")
                    text.insert(2, f"        <file_size>'{tar_file_size}'</file_size>")
                    text.insert(3, f"        <artist>'{self.submit_archive_artist_name_entry.text()}'</artist>")
                    text.insert(4, f"        <description>'{description_text}'</description>")
                    text.insert(5, '    </archive>')

                    out_file = open(xml_file, 'w')
                    for line in text:
                        print(line, file=out_file)
                    out_file.close()

                    return True

                def upload_archive():

                    self.ftp_upload_connect()

                    # Check to see if file already exists on ftp
                    ftp_file_list = self.ftp.nlst('/Submit_Archives')

                    if self.tar_file_name + '.tgz' in ftp_file_list:
                        QtWidgets.QApplication.restoreOverrideCursor()
                        self.ftp.quit()
                        PyFlameMessageWindow(
                            message='Archive already exists. Rename and try again.',
                            script_name=SCRIPT_NAME,
                            type=MessageType.ERROR
                        )
                        return

                    pyflame.message_print(
                        message='Uploading archive...',
                        script_name=SCRIPT_NAME,
                        )

                    # Close window
                    self.submit_archive_window.close()

                    # Upload tgz file to ftp
                    self.upload_file(
                        upload_type='Archive',
                        file_name=self.tar_file_name,
                        xml_path=xml_file,
                        tgz_path=self.tar_path,
                        )

                if not os.path.isdir(self.submit_archive_path_entry.text()):
                    PyFlameMessageWindow(
                        message='Enter valid archive path.',
                        script_name=SCRIPT_NAME,
                        type=MessageType.ERROR
                        )
                    return

                seg_file = [f for f in os.listdir(self.submit_archive_path_entry.text()) if f.endswith('.seg')]

                if not seg_file:
                    PyFlameMessageWindow(
                        message='Archive not found in selected folder.',
                        script_name=SCRIPT_NAME,
                        type=MessageType.ERROR
                        )
                    return
                if self.submit_archive_artist_name_entry.text() == '':
                    PyFlameMessageWindow(
                        message='Enter Artist name.',
                        script_name=SCRIPT_NAME,
                        type=MessageType.ERROR
                        )
                    return
                elif self.submit_archive_description_text_edit.toPlainText() == '':
                    PyFlameMessageWindow(
                        message='Enter archive description.',
                        script_name=SCRIPT_NAME,
                        type=MessageType.ERROR
                        )
                    return
                else:
                    if PyFlameMessageWindow(
                        message='All files in selected folder will be uploaded.',
                        script_name=SCRIPT_NAME,
                        type=MessageType.WARNING
                        ):
                        save_config()
                        compress_archive()
                        xml_file = os.path.join(self.temp_folder, f'{self.tar_file_name}.xml')
                        create_archive_xml_file()
                        if create_archive_xml_file():
                            upload_archive()

            flame_version = str(self.flame_version)
            if flame_version.endswith('.0'):
                flame_version = flame_version[:-2]

            # Create Archive Submit Window
            self.submit_archive_window = PyFlameWindow(
                width=800,
                height=530,
                title='Submit Flame Archive',
                return_pressed=archive_upload
                )

            # Labels
            self.submit_archive_label = PyFlameLabel(
                text='Logik Portal Archive Submit',
                style=Style.UNDERLINE,
                max_width=True,
                )
            self.submit_archive_path_label = PyFlameLabel(
                text='Archive Path',
                width=110,
                )
            self.submit_archive_name_label = PyFlameLabel(
                text='Archive Name',
                width=110,
                )
            self.submit_archive_flame_version_label = PyFlameLabel(
                text='Flame Version',
                width=110,
                )
            self.submit_archive_artist_name_label = PyFlameLabel(
                text='Artist Name',
                width=110,
                )
            self.submit_archive_description_label = PyFlameLabel(
                text='Description',
                width=110,
                )

            # Entries
            self.submit_archive_name_field = PyFlameLineEdit(
                text='',
                max_width=True,
                read_only=True,
                )
            self.submit_archive_flame_version_field = PyFlameLineEdit(
                text=flame_version,
                max_width=True,
                read_only=True,
                )
            self.submit_archive_path_entry = PyFlameLineEdit(
                text=self.settings.archive_submit_path,
                max_width=True,
                )
            self.submit_archive_artist_name_entry = PyFlameLineEdit(
                text='',
                max_width=True,
                )
            get_archive_name()

            # Line Edit File Browser
            self.submit_archive_path_entry = PyFlameLineEditFileBrowser(
                text=self.settings.archive_submit_path,
                max_width=True,
                connect=get_archive_info,
                browser_type=BrowserType.DIRECTORY,
                browser_title='Select Archive Folder',
                browser_window_to_hide=[self.window, self.submit_archive_window]
                )

            # TextEdit
            self.submit_archive_description_text_edit = PyFlameTextEdit(
                text=self.file_description,
                max_width=True,
                max_height=True,
                read_only=False,
                )

            # Buttons
            self.submit_archive_upload_btn = PyFlameButton(
                text='Upload',
                connect=archive_upload,
                color=Color.BLUE,
                )
            self.submit_archive_cancel_btn = PyFlameButton(
                text='Cancel',
                connect=self.submit_archive_window.close,
                )

            # Window Layout
            grid_layout = QtWidgets.QGridLayout()
            grid_layout.setColumnMinimumWidth(2, 150)
            grid_layout.setColumnMinimumWidth(3, 150)
            grid_layout.setRowMinimumHeight(11, 30)

            grid_layout.addWidget(self.submit_archive_label, 0, 0, 1, 6)
            grid_layout.addWidget(self.submit_archive_path_label, 1, 0)
            grid_layout.addWidget(self.submit_archive_name_label, 2, 0)
            grid_layout.addWidget(self.submit_archive_artist_name_label, 3, 0)
            grid_layout.addWidget(self.submit_archive_flame_version_label, 4, 0)
            grid_layout.addWidget(self.submit_archive_description_label, 5, 0)

            grid_layout.addWidget(self.submit_archive_path_entry, 1, 1, 1, 5)
            grid_layout.addWidget(self.submit_archive_name_field, 2, 1, 1, 5)
            grid_layout.addWidget(self.submit_archive_artist_name_entry, 3, 1, 1, 5)
            grid_layout.addWidget(self.submit_archive_flame_version_field, 4, 1, 1, 5)
            grid_layout.addWidget(self.submit_archive_description_text_edit, 5, 1, 6, 5)

            grid_layout.addWidget(self.submit_archive_cancel_btn, 12, 4)
            grid_layout.addWidget(self.submit_archive_upload_btn, 12, 5)

            # Add layout to main window
            self.submit_archive_window.add_layout(grid_layout)
            self.submit_archive_window.show()

        #  Archive Tab 5

        # Labels
        self.archives_label = PyFlameLabel(
            text='Archives',
            style=Style.UNDERLINE,
            max_width=True,
            )
        self.archives_desciption_label = PyFlameLabel(
            text='Archive Description',
            style=Style.UNDERLINE,
            max_width=True,
            )

        # Logo
        self.logik_logo_label = QtWidgets.QLabel()
        self.logo_label_pixmap = QtGui.QPixmap(os.path.join(SCRIPT_PATH, 'logik_logo.png'))
        self.logik_logo_label.setPixmap(self.logo_label_pixmap)

        # Text Edit
        self.archives_text_edit = PyFlameTextEdit(
            text=self.file_description,
            height=200,
            max_width=True,
            read_only=True,
            )

        # Batch Setups TreeWidget
        archive_tree_columns = [
            'Name',
            'Flame',
            'Size',
            'Artist'
            ]
        self.archives_tree = PyFlameTreeWidget(
            column_names=archive_tree_columns,
            connect=self.get_archive_description,
            height=250,
            max_width=True,
            )

        self.archives_tree.setColumnWidth(0, 600)
        self.archives_tree.setColumnWidth(1, 100)
        self.archives_tree.setColumnWidth(2, 100)
        self.archives_tree.setColumnWidth(3, 300)
        self.archives_tree.setTextElideMode(QtCore.Qt.ElideNone)

        # Disable batch download button if current flame version older than batch minimum
        self.archives_tree.clicked.connect(
            partial(
                self.check_archive_flame_version,
                self.archives_tree,
                )
            )

        # Buttons
        self.archives_submit_btn = PyFlameButton(
            text='Submit',
            connect=submit_archive,
            )
        self.archives_download_btn = PyFlameButton(
            text='Download',
            connect=archive_download,
            color=Color.BLUE,
            )
        self.archives_done_btn = PyFlameButton(
            text='Done',
            connect=self.done,
            )

        #------------------------------------#

        #  Tab layout
        self.window.tab5.layout = QtWidgets.QGridLayout()

        self.window.tab5.layout.addWidget(self.archives_label, 0, 0, 1, 5)

        self.window.tab5.layout.addWidget(self.archives_tree, 1, 0, 1, 5)

        self.window.tab5.layout.addWidget(self.archives_submit_btn, 2, 0)
        self.window.tab5.layout.addWidget(self.archives_download_btn, 2, 4)

        self.window.tab5.layout.addWidget(self.archives_desciption_label, 4, 0, 1, 5)
        self.window.tab5.layout.addWidget(self.archives_text_edit, 5, 0, 1, 5)

        self.window.tab5.layout.addWidget(self.logik_logo_label, 6, 0)
        self.window.tab5.layout.addWidget(self.archives_done_btn, 6, 4)

        self.window.tab5.setLayout(self.window.tab5.layout)

    # ----------------------------------------------------------------- #

    def get_updates(self):
        """
        Get updates.txt from FTP server and display in the python descriptions text edit when the Portal first opens.
        """

        def apply_update(label, text_edit):

            # Set the script description label to Logik Portal Updates
            label.setText('Logik Portal Updates')

            # Add updates to description text edit
            text_edit.setPlainText(self.updates)

        # Get the updates from the FTP server
        lines = []
        self.ftp.retrlines('RETR /Updates/updates.txt', lines.append)
        self.updates = '\n'.join(lines)
        # print('updates: ', self.updates)

        # Set the updates text edit to the updates
        if self.settings.last_tab == 0:
            apply_update(self.script_description_label, self.script_description_text_edit)
        elif self.settings.last_tab == 1:
            apply_update(self.matchbox_desciption_label, self.matchbox_text_edit)
        elif self.settings.last_tab == 2:
            apply_update(self.batch_setups_desciption_label, self.batch_setups_text_edit)
        elif self.settings.last_tab == 3:
            apply_update(self.archives_desciption_label, self.archives_text_edit)

    def clear_submit_script_ui(self):

        self.settings.script_submit_path = self.submit_script_path_entry.text()

        if os.path.isfile(self.settings.script_submit_path):
            self.submit_script_path_entry.setText(self.settings.script_submit_path)

            # Clear submit fields
            self.submit_script_version_lineedit.setText('')
            self.submit_script_flame_version_lineedit.setText('')
            self.submit_script_date_lineedit.setText('')
            self.submit_script_dev_name_lineedit.setText('')
            self.submit_script_description_text_edit.setPlainText('')

            self.get_script_info()

    def get_script_info(self):

        def extract_first_docstring(filename):
            """
            Extracts the first docstring from a python file and returns it as a string.
            Checks for both types of triple quote docstring formats.
            """

            with open(filename, 'r') as f:
                content = f.read()

            # Regular expression to match docstrings
            pattern = re.compile(r'(\'\'\'(.*?)\'\'\'|\"\"\"(.*?)\"\"\")', re.DOTALL)
            match = pattern.search(content)

            if match:
                return match.group(2) if match.group(2) else match.group(3)

            return ''

        # Get script description from script
        script_path = self.submit_script_path_entry.text()
        file_description = extract_first_docstring(script_path)
        file_description = file_description.strip()
        # print('file_description:', file_description)

        # Fill in script name field
        script_name = self.submit_script_path_entry.text().rsplit('/', 1)[1][:-3]
        script_name = script_name.replace('_', ' ')
        self.submit_script_name_field.setText(script_name)
        # print('script_name:', script_name)

        # Fill submit fields if data in present in script
        file_description_lines = file_description.splitlines()
        for line in file_description_lines:
            if 'Script Version: ' in line:
                script_version = line.split('Script Version: ', 1)[1]
                self.submit_script_version_lineedit.setText(script_version)
                # print('script_version:', script_version)
            elif 'Flame Version: ' in line:
                flame_version = line.split('Flame Version: ', 1)[1]
                self.submit_script_flame_version_lineedit.setText(flame_version)
                # print('flame_version:', flame_version)
            elif 'Written by: ' in line:
                script_dev = line.split('Written by: ', 1)[1]
                self.submit_script_dev_name_lineedit.setText(script_dev)
                # print('script_dev:', script_dev)
            elif 'Creation Date: ' in line:
                script_date = line.split('Creation Date: ', 1)[1]
            elif 'Update Date: ' in line:
                script_date = line.split('Update Date: ', 1)[1]
                self.submit_script_date_lineedit.setText(script_date)
                # print('script_date:', script_date, '\n')

        self.submit_script_description_text_edit.setText(file_description)

    def update_script_info(self):

        # Fill in Script info fields once valid script path is entered. Must end with .py
        if self.submit_script_path_entry.text().endswith('.py'):
            self.get_script_info()
        else:
            self.submit_script_name_field.setText('')
            self.submit_script_version_label_02.setText('')
            self.submit_script_flame_version_label_02.setText('')
            self.submit_script_date_label_02.setText('')
            self.submit_script_dev_name_label_02.setText('')
            self.submit_script_description_text_edit.setPlainText('')

    def installed_script_description(self):

        # Read description from script
        selected_script = self.installed_scripts_tree.selectedItems()
        script_item = selected_script[0]
        script_path = script_item.text(5)

        with open(script_path, 'r') as script:
            script_lines = script.read()

        try:
            file_description = re.split(r'"""|\'\'\'', script_lines, maxsplit=1)[1]
            file_description = re.split(r'"""|\'\'\'', file_description, maxsplit=1)[0]
            file_description = file_description.strip()
        except IndexError:
            file_description = ''

        self.script_description_text_edit.setPlainText(file_description)

    def get_installed_scripts(self, tree, scripts_root_path):

        # Clear tree list
        tree.clear()

        for root, dirs, files in os.walk(scripts_root_path, followlinks=True):
            if root[len(scripts_root_path):].count(os.sep) < 2:
                for script in files:
                    if script.endswith('.py'):
                        if not script.startswith(('.', 'flame_widgets', 'pyflame_lib')):

                            # Get script name from .py file name

                            script_name = script[:-3]
                            script_name = script_name.replace('_', ' ')
                            #print('script_name:', script_name)

                            script_path = os.path.join(root, script)
                            # print('script_path:', script_path)

                            # Read in script to separate out comments
                            script_code = open(script_path, 'r')
                            script_lines = script_code.read().splitlines()[1:]
                            script_code.close()

                            # Split out script info to comment list
                            comment_lines = []

                            for line in script_lines:
                                if line != '':
                                    comment_lines.append(line)
                                else:
                                    break

                            # Get script info from comment list
                            try:
                                script_version = [line.split('Script Version: ', 1)[1] for line in comment_lines if 'Script Version: ' in line]
                                if script_version:
                                    script_version = script_version[0]
                                else:
                                    script_version = [line.split("'", 2)[1] for line in script_lines if 'SCRIPT_VERSION = ' in line] # For old scripts
                                    if script_version:
                                        script_version = script_version[0]
                                        if 'v' in script_version:
                                            script_version = script_version[1:]
                                    else:
                                        script_version = ''
                            except:
                                script_version = ''

                            try:
                                script_date = [line.split('Update Date: ', 1)[1] for line in comment_lines if 'Update Date: ' in line]
                                if script_date:
                                    script_date = script_date[0]
                                else:
                                    script_date = [line.split(' ', 1)[1] for line in comment_lines if 'Updated:' in line] # For old scripts
                                    if script_date:
                                        script_date = script_date[0]
                                    else:
                                        script_date = ''
                            except:
                                script_date = ''

                            if script_date != '':
                                script_date = self.date_flip(script_date)

                            # print('script_date:', script_date)

                            try:
                                script_dev = [line.split('Written by: ', 1)[1] for line in comment_lines if 'Written by' in line]
                                if script_dev:
                                    script_dev = script_dev[0]
                                else:
                                    script_dev = [line.split('Created by ', 1)[1] for line in comment_lines if 'Created by' in line] # For old scripts
                                    if script_dev:
                                        script_dev = script_dev[0]
                                    else:
                                        script_dev = ''
                            except:
                                script_dev = ''
                            # print('script_dev:', script_dev)

                            try:
                                script_flame_version = [line.split('Flame Version: ', 1)[1] for line in comment_lines if 'Flame Version: ' in line]
                                if script_flame_version:
                                    script_flame_version = script_flame_version[0].split(' ', 1)[0]
                                else:
                                    script_flame_version = [line.split(' ', 1)[1] for line in comment_lines if 'Flame 20' in line] # For old scripts
                                    if script_flame_version:
                                        script_flame_version = script_flame_version[0].split(' ', 1)[0]
                                    else:
                                        script_flame_version = ''
                            except:
                                script_version = ''
                            # print('script_min_flame_version:', script_flame_version)

                            # Add script to tree

                            QtWidgets.QTreeWidgetItem(tree, [script_name, script_version, script_flame_version, script_date, script_dev, script_path])

                            self.installed_script_dict.update({script_name : script_version})

        # Set width of tree headers

        self.installed_scripts_tree.resizeColumnToContents(0)
        self.installed_scripts_tree.resizeColumnToContents(4)
        self.installed_scripts_tree.resizeColumnToContents(5)
        self.installed_scripts_tree.header().setSectionResizeMode(0, QtWidgets.QHeaderView.Fixed)
        self.installed_scripts_tree.header().setSectionResizeMode(1, QtWidgets.QHeaderView.Fixed)
        self.installed_scripts_tree.header().setSectionResizeMode(2, QtWidgets.QHeaderView.Fixed)
        self.installed_scripts_tree.header().setSectionResizeMode(3, QtWidgets.QHeaderView.Fixed)
        self.installed_scripts_tree.header().setSectionResizeMode(4, QtWidgets.QHeaderView.Fixed)
        self.installed_scripts_tree.header().setSectionResizeMode(5, QtWidgets.QHeaderView.Fixed)

        self.installed_scripts_tree.setTextElideMode(QtCore.Qt.ElideNone)

        # Select first item in tree

        tree.setCurrentItem(tree.topLevelItem(0))

    def delete_script(self, tree, root_script_path):

        # Get script path

        selected_script = tree.selectedItems()
        script_item = selected_script[0]
        script_to_delete = script_item.text(0)
        script_to_delete = script_to_delete.replace(' ', '_')
        script_path = script_item.text(5)
        script_folder_name = script_path.rsplit('/', 2)[1]
        script_folder_path = script_path.rsplit('/', 1)[0]

        # Confirm deletion
        if not PyFlameMessageWindow(
            message=f'Delete python script: <b>{script_to_delete}</b>',
            script_name=SCRIPT_NAME,
            type=MessageType.WARNING,
            ):
            pyflame.message_print(
                message='Delete cancelled.',
                script_name=SCRIPT_NAME,
                )
        else:

            # Remove script
            # If script is in folder with the same name of script, remove the folder

            if script_folder_name == script_to_delete:
                shutil.rmtree(script_folder_path)
            else:
                os.remove(script_path)
                try:
                    os.remove(script_path + 'c')
                except:
                    pass

            pyflame.message_print(
                message=f'Python script deleted: {script_to_delete}',
                script_name=SCRIPT_NAME,
                )

            # Update list of installed scripts

            self.get_installed_scripts(tree, root_script_path)


    # ----------------------------------------------------------------- #
    # Python Scripts

    def date_flip(self, date):
        """
        Swap date from mm.dd.yy to yy.mm.dd for python scripts list sorting.
        """

        date = date.split('.')
        date = date[2] + '.' + date[0] + '.' + date[1]

        return date

    def check_script_flame_version(self, tree, tree_index) -> None:
        """
        Check if script is compatible with current version of Flame.
        If script won't work with current version of Flame, disable the install button.

        Args:
            tree: TreeWidget
            tree_index: Index of selected item in tree. (Not used) Delete this at some point.
        """

        selected_script = tree.selectedItems()
        script_item = selected_script[0]
        script_name = script_item.text(0)
        script_flame_version = script_item.text(2)
        script_flame_max_version = script_item.text(3)

        if float(self.flame_version) < float(script_flame_version):
            print(f'--> {script_name}: Requires newer version of Flame.\n')
            self.install_script_btn.setEnabled(False)
        if script_flame_max_version != 'Latest':
            if float(script_flame_max_version) < float(self.flame_version):
                print(f'--> {script_name}: Does not work with this version of Flame.\n')
                self.install_script_btn.setEnabled(False)
        else:
            print(f'--> {script_name}: Flame version compatible.\n')
            self.install_script_btn.setEnabled(True)

    def get_script_description(self):

        print('--> Getting python script description.\n')

        # Switch text edit label to 'Script Description'. This is named 'Logik Portal Updates' when the script first loads.
        self.script_description_label.setText('Python Script Description')

        # Get selected script info
        selected_item = self.portal_scripts_tree.selectedItems()
        script = selected_item[0]
        script_name = script.text(0)
        author = script.text(4)

        # If script is not an Autodesk script, get description from ftp.
        # If script is an Autodesk script, get description from local file.
        if author != 'Autodesk':
            # Get script description from ftp
            xml_tree = ET.parse(self.python_scripts_xml_path)
            root = xml_tree.getroot()

            for script in root.findall('script'):
                if script.get('name') == script_name:
                    self.script_description_text_edit.setPlainText(script[-1].text[1:-1])

        elif author == 'Autodesk':
            # Get script description from local Autodesk file
            script_path = os.path.join(self.autodesk_scripts_path, script_name.replace(' ', '_') + '.py')

            with open(script_path, 'r') as f:
                script_text = f.read()
                script_text = f'Script Name: {script_name}\nFlame Version: {self.flame_version}\nWritten by: Autodesk\n\nDescription:\n' + script_text.split('"""')[1]
                print(script_text)

                self.script_description_text_edit.setPlainText(script_text)

    def get_ftp_scripts(self):
        """
        Add Logik Portal python scripts to Portal Python Scripts tree.

        Get script info from xml file and add to tree list. If a newer version of the script exists on the ftp, highlight the script entry.
        If the script requires a newer version of flame, grey out the script entry.

        """

        # Add items from xml to scripts tree list
        xml_tree = ET.parse(self.python_scripts_xml_path)
        root = xml_tree.getroot()

        # Extract values for each script from xml. If max flame version not present, set to '-'.
        for script_elem in root.findall('./script'):
            script_name = script_elem.get('name')
            script_version = script_elem.find('script_version').text.strip("'")
            flame_min_version = script_elem.find('flame_version').text.strip("'")
            try:
                flame_max_version = script_elem.find('flame_max_version').text.strip("'")
            except:
                flame_max_version = 'Latest'
            date = script_elem.find('date').text.strip("'")
            developer_name = script_elem.find('developer').text.strip("'")

            # print("Script Name:", script_name)
            # print("Script Version:", script_version)
            # print("Flame Min Version:", flame_min_version)
            # print("Flame Max Version:", flame_max_version)
            # print("Date:", date)
            # print("Developer:", developer_name)
            # print()

            # Add script info to tree list
            new_script = QtWidgets.QTreeWidgetItem(self.portal_scripts_tree, [script_name, script_version, flame_min_version, flame_max_version, date, developer_name])

            # If newer version of script exists on ftp, highlight script entry
            if script_name in self.installed_script_dict:
                installed_script_version = self.installed_script_dict.get(script_name)
                try:
                    if float(script_version) > float(installed_script_version):
                        new_script.setForeground(0, QtGui.QColor('#ffffff'))
                        new_script.setForeground(1, QtGui.QColor('#ffffff'))
                        new_script.setForeground(2, QtGui.QColor('#ffffff'))
                        new_script.setForeground(3, QtGui.QColor('#ffffff'))
                        new_script.setForeground(4, QtGui.QColor('#ffffff'))
                        new_script.setForeground(5, QtGui.QColor('#ffffff'))
                except:
                    pass

            # if script requires newer version of flame grey out script entry
            if float(self.flame_version) < float(flame_min_version):
                new_script.setForeground(0, QtGui.QColor('#555555'))
                new_script.setForeground(1, QtGui.QColor('#555555'))
                new_script.setForeground(2, QtGui.QColor('#555555'))
                new_script.setForeground(3, QtGui.QColor('#555555'))
                new_script.setForeground(4, QtGui.QColor('#555555'))
                new_script.setForeground(5, QtGui.QColor('#555555'))

            # If scripts max_flame_version if not equal to the current flame version or 'Latest', grey out the script entry.
            if flame_max_version != 'Latest':
                if float(flame_max_version) < float(self.flame_version):
                    new_script.setForeground(0, QtGui.QColor('#555555'))
                    new_script.setForeground(1, QtGui.QColor('#555555'))
                    new_script.setForeground(2, QtGui.QColor('#555555'))
                    new_script.setForeground(3, QtGui.QColor('#555555'))
                    new_script.setForeground(4, QtGui.QColor('#555555'))
                    new_script.setForeground(5, QtGui.QColor('#555555'))

            self.ftp_script_list.append(script_name)

        # Select top item in script setup list
        self.portal_scripts_tree.setCurrentItem(self.portal_scripts_tree.topLevelItem(0))

        # Get selected script setup description
        self.get_script_description()

        # Set width of tree headers
        self.portal_scripts_tree.resizeColumnToContents(0)
        self.portal_scripts_tree.resizeColumnToContents(4)
        self.portal_scripts_tree.header().setSectionResizeMode(0, QtWidgets.QHeaderView.Fixed)
        self.portal_scripts_tree.header().setSectionResizeMode(1, QtWidgets.QHeaderView.Fixed)
        self.portal_scripts_tree.header().setSectionResizeMode(2, QtWidgets.QHeaderView.Fixed)
        self.portal_scripts_tree.header().setSectionResizeMode(3, QtWidgets.QHeaderView.Fixed)
        self.portal_scripts_tree.header().setSectionResizeMode(4, QtWidgets.QHeaderView.Fixed)
        self.portal_scripts_tree.header().setSectionResizeMode(5, QtWidgets.QHeaderView.Fixed)

        self.portal_scripts_tree.setTextElideMode(QtCore.Qt.ElideNone)

    def get_autodesk_scripts(self):
        """
        Add Autodesk scripts to script setup list from: /opt/Autodesk/FLAME SCRIPT_VERSION/python_utilities/scripts
        For Flame 2023.2+ only.
        """

        if self.flame_version >= 2023.2:

            print('--> Adding Autodesk python scripts.\n')

            for script in os.listdir(self.autodesk_scripts_path):
                script_name = script.rsplit('.', 1)[0].replace('_', ' ')
                script_version = '1.0'
                flame_version = str(self.flame_version)
                flame_max_version = str(self.flame_version)
                developer_name = 'Autodesk'

                timestamp = os.path.getmtime(os.path.join(self.autodesk_scripts_path, script))
                datestamp = datetime.datetime.fromtimestamp(timestamp)
                date = datestamp.strftime("%y.%m.%d")

                QtWidgets.QTreeWidgetItem(self.portal_scripts_tree, [script_name, script_version, flame_version, flame_max_version, date, developer_name])

    # ----------------------------------------------------------------- #
    # Common

    def get_description(self, label, label_text, tree, xml_path, findall, text_edit, index):
        """
        Get item description from xml file and show in text edit.
        """

        print(f'--> Getting {label_text}.\n')

        selected_item = tree.selectedItems()
        item = selected_item[0]
        item_name = item.text(0)
        item_name = item_name.replace(' ', '_')

        # Add items from xml to matchbox list

        xml_tree = ET.parse(xml_path)
        root = xml_tree.getroot()

        for item in root.findall(findall):
            if item.get('name') == item_name:
                text_edit.setPlainText(item[index].text[1:-1])

        # Switch text edit label to 'Script Description'. This is named 'Logik Portal Updates' when the script first loads.

        label.setText(label_text)

    def download_xmls(self) -> None:
        """
        Download file list xmls from ftp to temp folder.
        """

        print(f'--> Downloading XML files...\n')

        # Download xmls to temp folder

        self.python_scripts_xml_path = os.path.join(self.temp_folder, 'python_scripts.xml')
        self.ftp.retrbinary('RETR ' + '/Scripts/python_scripts.xml', open(self.python_scripts_xml_path, 'wb').write)

        self.matchbox_xml_path = os.path.join(self.temp_folder, 'matchbox_collection.xml')
        self.ftp.retrbinary('RETR ' + '/Logik_Matchbox/matchbox_collection.xml', open(self.matchbox_xml_path, 'wb').write)

        self.batch_setups_xml_path = os.path.join(self.temp_folder, 'batch_setups.xml')
        self.ftp.retrbinary('RETR ' + '/Batch_Setups/batch_setups.xml', open(self.batch_setups_xml_path, 'wb').write)

        self.archives_xml_path = os.path.join(self.temp_folder, 'archives.xml')
        self.ftp.retrbinary('RETR ' + '/Archives/archives.xml', open(self.archives_xml_path, 'wb').write)

        print(f'--> XML files downloaded.\n')

    def download_file(self, download_type: str, file_name: str, ftp_file_path: str, tgz_path: str) -> None:
        """
        Download selected file from ftp with progress window

        Parameters:
            download_type (str): Type of file being downloaded (Python Script, Matchbox, Batch Setup, Archive)
            file_name (str): Name of file being downloaded.
            ftp_file_path (str): Path to file on ftp.
            tgz_path (str): Path to download file to.
        """

        # Connect to ftp
        self.ftp_download_connect()

        # Get file size
        file_size = self.ftp.size(ftp_file_path)

        file_size = file_size / 1000
        file_size = int(round(file_size, 1))
        num_to_do = file_size
        file_size = str(file_size) + ' KB'

        # Create progress window
        self.progress_window = PyFlameProgressWindow(
            num_to_do=num_to_do,
            title=f'{SCRIPT_NAME}: Downloading',
            text=f'{download_type}: {file_name}<br><br>0 KB of {file_size}',
            )

        pyflame.message_print(
            message=f'Downloading {download_type}: {file_name}',
            script_name=SCRIPT_NAME,
            )

        # Variables to store progress
        downloaded = 0

        # Callback function to write data and update progress
        def write_and_update(data):
            nonlocal downloaded

            with open(tgz_path, 'ab') as f:
                f.write(data)

            downloaded += len(data)
            downloaded_bytes = downloaded / 1000
            downloaded_progress = int(round(downloaded_bytes, 1))
            downloaded_bytes = str(downloaded_progress) + ' KB'
            self.progress_window.set_progress_value(downloaded_progress)
            self.progress_window.set_text(f'{download_type}: {file_name}<br><br>{downloaded_bytes} of {file_size}')

        # Retrieve file in binary mode with a callback
        self.ftp.retrbinary(f'RETR {ftp_file_path}', write_and_update)

        # Enable done button on progress window when download is complete
        self.progress_window.set_text(f'{download_type}: {file_name}<br><br>{file_size} of {file_size}')
        self.progress_window.set_progress_value(num_to_do)
        self.progress_window.enable_done_button(True)

        # Disconnect from ftp
        self.ftp.quit()

    def upload_file(self, upload_type: str, file_name: str, xml_path: str, tgz_path: str) -> None:
        """
        Upload selected file to ftp with progress window

        Parameters:
            upload_type (str): Type of file being uploaded (Python Script, Matchbox, Batch Setup, Archive)
            file_name (str): Name of file being uploaded.
            xml_path (str): Path to xml file to be uploaded to ftp.
            tgz_path (str): Path to tgz file to be uploaded to ftp.
        """

        QtWidgets.QApplication.setOverrideCursor(QtCore.Qt.BusyCursor)

        pyflame.message_print(
            message=f'Uploading: {file_name}',
            script_name=SCRIPT_NAME,
            )

        if upload_type == 'Python Script':
            upload_folder = '/Submit_Scripts'
        elif upload_type == 'Batch Setup':
            upload_folder = '/Submit_Batch_Setups'
        elif upload_type == 'Archive':
            upload_folder = '/Submit_Archives'

        # Upload archive tar to ftp
        ftp_tgz_path = os.path.join(upload_folder, file_name) + '.tgz'

        # Upload archive xml file
        ftp_xml_path = os.path.join(upload_folder, file_name) + '.xml'

        # Connect to ftp if not uploading python script - python script uses a different ftp connection
        if not upload_type == 'Python Script':
            self.ftp_upload_connect()

        # Get file size
        file_size = os.path.getsize(tgz_path)
        num_to_do = file_size

        # if the number of digits in file_size is greater than 7 then divide by 1000000 to get MB
        if len(str(file_size)) > 7:
            file_size = file_size / 1000
            num_to_do = int(round(file_size, 1))
            file_size = str(num_to_do) + ' KB'

        # Create progress window
        self.progress_window = PyFlameProgressWindow(
            num_to_do=num_to_do,
            title=f'{SCRIPT_NAME}: Uploading',
            text=f'{upload_type}: {file_name}<br><br>0 KB of {file_size}',
            )

        # Upload xml file to ftp
        with open(xml_path, 'rb') as f:
            self.ftp.storbinary(f'STOR {ftp_xml_path}', f)
        print(f'--> {file_name}.xml uploaded to ftp.\n')

        # Upload tgz file to ftp

        # Variables to store progress
        uploaded = 0

        # Callback function to read data and update progress
        def read_and_update(data):
            nonlocal uploaded

            uploaded += len(data)
            uploaded_bytes = uploaded / 1000
            upload_progress = int(round(uploaded_bytes, 1))
            uploaded_bytes = str(upload_progress) + ' KB'

            self.progress_window.set_progress_value(upload_progress)
            self.progress_window.set_text(f'{upload_type}: {file_name}<br><br>{uploaded_bytes} of {file_size}')
            return data

        # Open the file and upload in binary mode with a callback
        with open(tgz_path, 'rb') as f:
            self.ftp.storbinary(f'STOR {ftp_tgz_path}', f, callback=read_and_update)

        QtWidgets.QApplication.restoreOverrideCursor()

        # Enable done button on progress window when upload is complete
        self.progress_window.set_text(f'{upload_type}: {file_name}<br><br>{file_size} {bytes} of {file_size} {bytes}')
        self.progress_window.set_progress_value(num_to_do)
        self.progress_window.enable_done_button(True)

        # Check that both files were uploaded to site
        ftp_file_list = self.ftp.nlst(upload_folder)

        if f'{os.path.join(upload_folder, file_name)}.tgz' and f'{os.path.join(upload_folder, file_name)}.xml' not in ftp_file_list:
            QtWidgets.QApplication.restoreOverrideCursor()
            self.ftp.quit()
            PyFlameMessageWindow(
                message='Upload failed. Try again.',
                script_name=SCRIPT_NAME,
                type=MessageType.ERROR
            )
            return

        # Clean up temp folder
        os.remove(xml_path)
        os.remove(tgz_path)

        # Disconnect from ftp
        self.ftp.quit()

    # ----------------------------------------------------------------- #
    # Matchbox

    def get_matchbox_description(self):

            self.get_description(
                self.matchbox_desciption_label,
                'Matchbox Description',
                self.matchbox_tree,
                self.matchbox_xml_path,
                'matchbox',
                self.matchbox_text_edit,
                5
            )

    def update_matchbox_tree(self, search: str=None) -> None:
        """
        search: string to search for in matchbox name
        """

        print('--> Updating matchbox list.\n')

        def add_matchbox(matchbox):
            shader_type = str(matchbox[0].text[1:-1])
            author_name = str(matchbox[4].text[1:-1])
            matchbox = QtWidgets.QTreeWidgetItem(self.matchbox_tree, [matchbox_name, shader_type, author_name])

        self.matchbox_tree.clear()

        # Add items from xml to matchbox list

        xml_tree = ET.parse(self.matchbox_xml_path)
        root = xml_tree.getroot()

        if search:
            for matchbox in root.findall('matchbox'):
                matchbox_name = str(matchbox.get('name'))
                if search.lower() in matchbox_name.lower():
                    add_matchbox(matchbox)
        else:
            for matchbox in root.findall('matchbox'):
                matchbox_name = str(matchbox.get('name'))
                add_matchbox(matchbox)

        # Select top item in matchbox list
        self.matchbox_tree.setCurrentItem(self.matchbox_tree.topLevelItem(0))

        # Get selected matchbox description
        self.get_matchbox_description()

    # ----------------------------------------------------------------- #
    # Batch Setups

    def check_batch_flame_version(self, tree, tree_index):

        print('--> Checking batch version.\n')

        # Get selected script date

        selected_batch = tree.selectedItems()
        batch_item = selected_batch[0]
        batch_name = batch_item.text(0)
        batch_flame_version = batch_item.text(1)

        # print('current_flame_version:', self.flame_version)
        # print('batch_flame_version:', batch_flame_version, '\n')

        if float(batch_flame_version) > float(self.flame_version):
            print(f'--> {batch_name} requires newer version of flame.\n')
            self.batch_setups_download_btn.setEnabled(False)
        else:
            self.batch_setups_download_btn.setEnabled(True)

    def get_batch_description(self):

        self.get_description(
            self.batch_setups_desciption_label,
            'Batch Setup Description',
            self.batch_setups_tree,
            self.batch_setups_xml_path,
            'batch',
            self.batch_setups_text_edit,
            2
        )

    def get_batch_setups(self):

        # Add items from xml to batch list

        xml_tree = ET.parse(self.batch_setups_xml_path)
        root = xml_tree.getroot()

        for batch in root.findall('batch'):
            batch_name = str(batch.get('name'))
            artist_name = str(batch[0].text[1:-1])
            flame_version = str(batch[1].text[1:-1])

            # print('batch_name:', batch_name)
            # print('artist_name:', artist_name)
            # print('flame_version:', flame_version, '\n')

            batch_setup = QtWidgets.QTreeWidgetItem(self.batch_setups_tree, [batch_name, flame_version, artist_name])

            # if batch setup requires newer version of flame grey out script entry

            if float(self.flame_version) < float(flame_version):
                batch_setup.setForeground(0, QtGui.QColor('#555555'))
                batch_setup.setForeground(1, QtGui.QColor('#555555'))
                batch_setup.setForeground(2, QtGui.QColor('#555555'))

        # Select top item in batch setup list

        self.batch_setups_tree.setCurrentItem(self.batch_setups_tree.topLevelItem(0))

        # Get selected batch setup description

        self.get_batch_description()

    # ----------------------------------------------------------------- #
    # Archives

    def check_archive_flame_version(self, tree, tree_index):

        print('--> Checking archive version.\n')

        # Get selected script date

        selected_archive = tree.selectedItems()
        archive_item = selected_archive[0]
        archive_name = archive_item.text(0)
        archive_flame_version = archive_item.text(1)

        # print('current_flame_version:', self.flame_version)
        # print('archive_flame_version:', archive_flame_version, '\n')

        if float(archive_flame_version) > float(self.flame_version):
            print(f'--> {archive_name} requires newer version of flame.\n')
            self.archives_download_btn.setEnabled(False)
        else:
            self.archives_download_btn.setEnabled(True)

    def get_archive_description(self):

        self.get_description(
            self.archives_desciption_label,
            'Archive Description',
            self.archives_tree,
            self.archives_xml_path,
            'archive',
            self.archives_text_edit,
            3
        )

    def get_archives(self):

        # Add items from xml to archive list
        xml_tree = ET.parse(self.archives_xml_path)
        root = xml_tree.getroot()

        for arc in root.findall('archive'):
            archive_name = str(arc.get('name'))
            flame_version = str(arc[0].text[1:-1])
            archive_size = str(arc[1].text[1:-1])
            artist_name = str(arc[2].text[1:-1])

            archive = QtWidgets.QTreeWidgetItem(self.archives_tree, [archive_name, flame_version, archive_size, artist_name])

            # if archive requires newer version of flame grey out script entry
            if float(self.flame_version) < float(flame_version):
                archive.setForeground(0, QtGui.QColor('#555555'))
                archive.setForeground(1, QtGui.QColor('#555555'))
                archive.setForeground(2, QtGui.QColor('#555555'))
                archive.setForeground(3, QtGui.QColor('#555555'))

        # Select top item in batch setup list
        self.archives_tree.setCurrentItem(self.archives_tree.topLevelItem(0))

        # Get selected batch setup description
        self.get_archive_description()

    # ----------------------------------------------------------------- #

    def done(self):

        # Save current tab to config
        self.settings.save_config(
            script_name=SCRIPT_NAME,
            script_path=SCRIPT_PATH,
            config_values={
                'last_tab': self.main_tabs.currentIndex()
                }
            )

        self.window.close()

        try:
            self.submit_script_window.close()
        except:
            pass

        try:
            self.submit_batch_window.close()
        except:
            pass

        try:
            self.ftp.quit()
        except:
            pass

        try:
            shutil.rmtree(self.temp_folder)
            print('--> Clearing temp files.\n')
        except:
            pass

        print('Done.\n')

# ---------------------------------------- #
# Flame Menus

def get_main_menu_custom_ui_actions():

    return [
        {
            'name': 'Logik',
            'actions': [
                {
                    'name': 'Logik Portal',
                    'execute': LogikPortal,
                    'minimumVersion': '2022',
                    'maximumVersion': '2023.1'
                }
            ]
        },
        {
            'name': 'Logik',
            'hierarchy': [],
            'actions': [
                {
                    'name': 'Logik Portal',
                    'order': 1,
                    'execute': LogikPortal,
                    'minimumVersion': '2023.2'
                }
            ]
        }
    ]
